# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
"""Redis-related providers."""
from __future__ import print_function

import codecs
import errno
import os
import subprocess
import sys
import time

from anaconda_project.requirements_registry.provider import (EnvVarProvider, ProviderAnalysis,
                                                             shutdown_service_run_state, delete_service_directory)
import anaconda_project.requirements_registry.network_util as network_util
from anaconda_project.provide import PROVIDE_MODE_DEVELOPMENT
from anaconda_project.frontend import _new_error_recorder
from anaconda_project.internal import py2_compat
from anaconda_project.internal import logged_subprocess

_DEFAULT_SYSTEM_REDIS_HOST = "localhost"
_DEFAULT_SYSTEM_REDIS_PORT = 6379
_DEFAULT_SYSTEM_REDIS_URL = "redis://%s:%d" % (_DEFAULT_SYSTEM_REDIS_HOST, _DEFAULT_SYSTEM_REDIS_PORT)


class _RedisProviderAnalysis(ProviderAnalysis):
    """Subtype of ProviderAnalysis with extra fields RedisProvider needs to track."""

    def __init__(self, config, missing_to_configure, missing_to_provide, existing_scoped_instance_url,
                 default_system_exists):
        super(_RedisProviderAnalysis, self).__init__(config, missing_to_configure, missing_to_provide)
        self.existing_scoped_instance_url = existing_scoped_instance_url
        self.default_system_exists = default_system_exists


# future: this should introduce a requirement that redis-server is on path
class RedisProvider(EnvVarProvider):
    """Runs a project-scoped Redis process (each project needing Redis gets its own)."""

    @classmethod
    def _parse_port_range(cls, s):
        pieces = s.split("-")
        if len(pieces) != 2:
            return None
        try:
            lower = int(pieces[0].strip())
            upper = int(pieces[1].strip())
        except ValueError:
            return None
        if lower <= 0 or upper <= 0:
            return None
        if lower > upper:
            return None
        return (lower, upper)

    def _config_section(self, requirement):
        return ["service_options", requirement.env_var]

    def read_config(self, requirement, environ, local_state_file, default_env_spec_name, overrides):
        """Override superclass to return our config."""
        config = super(RedisProvider, self).read_config(requirement, environ, local_state_file, default_env_spec_name,
                                                        overrides)

        assert 'source' in config

        section = self._config_section(requirement)

        scope = local_state_file.get_value(section + ['scope'], default='all')
        if config['source'] == 'unset':
            config['source'] = 'find_' + scope

        default_lower_port = 6380  # one above 6379 default Redis
        default_upper_port = 6449  # entirely arbitrary
        default_port_range = "%d-%d" % (default_lower_port, default_upper_port)
        port_range_string = local_state_file.get_value(section + ['port_range'], default=default_port_range)
        parsed_port_range = self._parse_port_range(port_range_string)
        if parsed_port_range is None:
            print(
                "Invalid port_range '%s', should be like '%s'" % (port_range_string, default_port_range),
                file=sys.stderr)
            config['lower_port'] = default_lower_port
            config['upper_port'] = default_upper_port
        else:
            config['lower_port'] = parsed_port_range[0]
            config['upper_port'] = parsed_port_range[1]

        return config

    def set_config_values_as_strings(self, requirement, environ, local_state_file, default_env_spec_name, overrides,
                                     values):
        """Override superclass to set our config values."""
        config = self.read_config(requirement, environ, local_state_file, default_env_spec_name, overrides=None)
        section = self._config_section(requirement)
        upper_port = config['upper_port']
        lower_port = config['lower_port']
        if 'lower_port' in values:
            lower_port = values['lower_port']
        if 'upper_port' in values:
            upper_port = values['upper_port']

        local_state_file.set_value(section + ['port_range'], "%s-%s" % (lower_port, upper_port))

        if 'source' in values:
            if values['source'] == 'find_all':
                scope = 'all'
            elif values['source'] == 'find_project':
                scope = 'project'
            elif values['source'] == 'find_system':
                scope = 'system'
            else:
                scope = None
            if scope is not None:
                local_state_file.set_value(section + ['scope'], scope)

            if values['source'] != 'environ':
                # clear out the previous setting; this is sort of a hack. The problem
                # is that we don't want to delete env vars set in actual os.environ on
                # the command line, in our first pass, and in some subtypes of EnvVarProvider
                # (CondaEnvProvider) we also don't want to use it by default. Otherwise
                # we should probably do this in EnvVarProvider. future: rethink this.
                # a possible fix is to track an initial_environ for the whole prepare
                # sequence, separately from the current running environ?
                environ.pop(requirement.env_var, None)

        # set a manually-specified value
        super(RedisProvider, self).set_config_values_as_strings(requirement, environ, local_state_file,
                                                                default_env_spec_name, overrides, values)

    def _previously_run_redis_url_if_alive(self, run_state):
        if 'port' in run_state and network_util.can_connect_to_socket(host='localhost', port=run_state['port']):
            return "redis://localhost:{port}".format(port=run_state['port'])
        else:
            return None

    def _can_connect_to_system_default(self):
        return network_util.can_connect_to_socket(host=_DEFAULT_SYSTEM_REDIS_HOST, port=_DEFAULT_SYSTEM_REDIS_PORT)

    def analyze(self, requirement, environ, local_state_file, default_env_spec_name, overrides):
        """Override superclass to store additional fields in the analysis."""
        analysis = super(RedisProvider, self).analyze(requirement, environ, local_state_file, default_env_spec_name,
                                                      overrides)
        run_state = local_state_file.get_service_run_state(requirement.env_var)
        previous = self._previously_run_redis_url_if_alive(run_state)
        systemwide = self._can_connect_to_system_default()

        return _RedisProviderAnalysis(
            analysis.config,
            analysis.missing_env_vars_to_configure,
            analysis.missing_env_vars_to_provide,
            existing_scoped_instance_url=previous,
            default_system_exists=systemwide)

    def _provide_system(self, requirement, context, frontend):
        if context.status.analysis.default_system_exists:
            frontend.info("Found system default Redis at %s" % _DEFAULT_SYSTEM_REDIS_URL)
            return _DEFAULT_SYSTEM_REDIS_URL

    def _provide_project(self, requirement, context, frontend):
        config = context.status.analysis.config

        def ensure_redis(run_state):
            # this is pretty lame, we'll want to get fancier at a
            # future time (e.g. use Chalmers, stuff like
            # that). The desired semantic is a new copy of Redis
            # dedicated to this project directory; it should not
            # require the user to have set up anything in advance,
            # e.g. if we use Chalmers we should automatically take
            # care of configuring/starting Chalmers itself.
            url = context.status.analysis.existing_scoped_instance_url
            if url is not None:
                frontend.info("Using redis-server we started previously at {url}".format(url=url))
                return url

            run_state.clear()

            workdir = context.ensure_service_directory(requirement.env_var)
            pidfile = os.path.join(workdir, "redis.pid")
            logfile = os.path.join(workdir, "redis.log")

            # 6379 is the default Redis port; leave that one free
            # for a systemwide Redis. Try looking for a port above
            # it. This is a pretty huge hack and a race condition,
            # but Redis doesn't as far as I know have "let the OS
            # pick the port" mode.
            LOWER_PORT = config['lower_port']
            UPPER_PORT = config['upper_port']
            port = LOWER_PORT
            while port <= UPPER_PORT:
                if not network_util.can_connect_to_socket(host='localhost', port=port):
                    break
                port += 1
            if port > UPPER_PORT:
                frontend.error(("All ports from {lower} to {upper} were in use, " +
                                "could not start redis-server on one of them.").format(
                                    lower=LOWER_PORT, upper=UPPER_PORT))
                return None

            # be sure we don't get confused by an old log file
            try:
                os.remove(logfile)
            except IOError:  # pragma: no cover (py3 only)
                pass
            except OSError:  # pragma: no cover (py2 only)
                pass

            command = [
                'redis-server', '--pidfile', pidfile, '--logfile', logfile, '--daemonize', 'yes', '--port',
                str(port)
            ]
            frontend.info("Starting " + repr(command))

            # we don't close_fds=True because on Windows that is documented to
            # keep us from collected stderr. But on Unix it's kinda broken not
            # to close_fds. Hmm.
            try:
                popen = logged_subprocess.Popen(
                    args=command, stderr=subprocess.PIPE, env=py2_compat.env_without_unicode(context.environ))
            except Exception as e:
                frontend.error("Error executing redis-server: %s" % (str(e)))
                return None

            # communicate() waits for the process to exit, which
            # is supposed to happen immediately due to --daemonize
            (out, err) = popen.communicate()
            assert out is None  # because we didn't PIPE it
            err = err.decode(errors='replace')

            url = None
            if popen.returncode == 0:
                # now we need to wait for Redis to be ready; we
                # are not sure whether it will create the port or
                # pidfile first, so wait for both.
                port_is_ready = False
                pidfile_is_ready = False
                MAX_WAIT_TIME = 10
                so_far = 0
                while so_far < MAX_WAIT_TIME:
                    increment = MAX_WAIT_TIME / 500.0
                    time.sleep(increment)
                    so_far += increment
                    if not port_is_ready:
                        if network_util.can_connect_to_socket(host='localhost', port=port):
                            port_is_ready = True

                    if not pidfile_is_ready:
                        if os.path.exists(pidfile):
                            pidfile_is_ready = True

                    if port_is_ready and pidfile_is_ready:
                        break

                # if we time out with no pidfile we forge ahead at this point
                if port_is_ready:
                    run_state['port'] = port
                    url = "redis://localhost:{port}".format(port=port)

                    # note: --port doesn't work, only -p, and the failure with --port is silent.
                    run_state['shutdown_commands'] = [['redis-cli', '-p', str(port), 'shutdown']]
                else:
                    frontend.info(
                        "redis-server started successfully, but we timed out trying to connect to it on port %d" %
                        (port))

            if url is None:
                for line in err.split("\n"):
                    if line != "":
                        frontend.info(line)
                try:
                    with codecs.open(logfile, 'r', 'utf-8') as log:
                        for line in log.readlines():
                            frontend.info(line)
                except IOError as e:
                    # just be silent if redis-server failed before creating a log file,
                    # that's fine. Hopefully it had some stderr.
                    if e.errno != errno.ENOENT:
                        frontend.info("Failed to read {logfile}: {error}".format(logfile=logfile, error=e))

                frontend.error(
                    "redis-server process failed or timed out, exited with code {code}".format(code=popen.returncode))

            return url

        return context.transform_service_run_state(requirement.env_var, ensure_redis)

    def provide(self, requirement, context):
        """Override superclass to start a project-scoped redis-server.

        If it locates or starts a redis-server, it sets the
        requirement's env var to that server's URL.

        """
        assert 'PATH' in context.environ

        url = None
        source = context.status.analysis.config['source']

        super_result = super(RedisProvider, self).provide(requirement, context)

        url = context.environ.get(requirement.env_var, None)

        frontend = _new_error_recorder(context.frontend)

        # we jump through a little hoop to avoid a "can't connect to system redis"
        # message if we're going to end up successfully starting a local one.
        system_failed = False
        if url is None and (source == 'find_system' or source == 'find_all'):
            url = self._provide_system(requirement, context, frontend)
            if url is None:
                system_failed = True

        if url is None and (source == 'find_project' or source == 'find_all'):
            # we will only start a local Redis in "dev" mode, not prod or check mode
            if context.mode == PROVIDE_MODE_DEVELOPMENT:
                url = self._provide_project(requirement, context, frontend)

        if url is None:
            if system_failed:
                frontend.error("Could not connect to system default Redis.")
        else:
            context.environ[requirement.env_var] = url

        return super_result.copy_with_additions(errors=frontend.pop_errors())

    def unprovide(self, requirement, environ, local_state_file, overrides, requirement_status=None):
        """Override superclass to shut down any redis-server we started."""
        status = shutdown_service_run_state(local_state_file, requirement.env_var)
        delete_service_directory(local_state_file, requirement.env_var)
        return status



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
from __future__ import absolute_import, print_function

from tornado.httpserver import HTTPServer
from tornado.netutil import bind_sockets
from tornado.web import Application, RequestHandler
from tornado import gen
import uuid
import hashlib
import socket


class _DownloadView(RequestHandler):
    def __init__(self, application, *args, **kwargs):
        # Note: application is stored as self.application
        super(_DownloadView, self).__init__(application, *args, **kwargs)

    @gen.coroutine
    def get(self, *args, **kwargs):
        download_id = self.get_argument("id")
        hash_algorithm = self.get_argument("hash_algorithm", None)
        length = int(self.get_argument("length"))

        print("Planning to send %d bytes" % length)
        if hash_algorithm:
            hasher = getattr(hashlib, hash_algorithm)()

        self.set_status(200)
        self.set_header('Content-Length', str(length))
        data = ("abcdefghijklmnop" * 20).encode("utf-8")
        remaining = length
        while remaining > 0:
            to_write = data[:remaining]
            if hash_algorithm:
                hasher.update(to_write)
            remaining = remaining - len(to_write)
            self.write(to_write)
            try:
                yield self.flush()
            except Exception as e:
                raise e

        if hash_algorithm:
            self.application.hashes[download_id] = hasher.hexdigest()

        self.finish()


class _ErrorView(RequestHandler):
    def __init__(self, application, *args, **kwargs):
        # Note: application is stored as self.application
        super(_ErrorView, self).__init__(application, *args, **kwargs)

    def get(self, *args, **kwargs):
        self.set_status(404)
        self.finish()


class _TestServerApplication(Application):
    def __init__(self, **kwargs):
        self.hashes = dict()
        patterns = [(r'/download', _DownloadView), (r'/error', _ErrorView)]
        super(_TestServerApplication, self).__init__(patterns, **kwargs)


class _TestServer(object):
    def __init__(self):
        self._application = _TestServerApplication()
        self._http = HTTPServer(self._application)

        # these would throw OSError on failure
        sockets = bind_sockets(port=None, address='127.0.0.1')

        self._port = None
        for s in sockets:
            # we have to find the ipv4 one
            if s.family == socket.AF_INET:
                self._port = s.getsockname()[1]
        assert self._port is not None

        self._http.add_sockets(sockets)

    @property
    def port(self):
        return self._port

    @property
    def url(self):
        return "http://localhost:%d/" % self.port

    def start(self):
        self._http.start(1)

    def unlisten(self):
        self._http.close_all_connections()
        self._http.stop()

    @property
    def error_url(self):
        return self.url + "error"

    def new_download_url(self, download_length, hash_algorithm):
        url = (self.url + "download?id=" + str(uuid.uuid4()) + "&length=" + str(download_length))
        if hash_algorithm:
            url += "&hash_algorithm=" + hash_algorithm
        return url

    def server_computed_hash_for_downloaded_url(self, download_url):
        i = download_url.index("id=")
        download_id = download_url[(i + 3):][:36]
        if download_id not in self._application.hashes:
            raise RuntimeError("It looks like the download from %s did not complete" % download_url)
        return self._application.hashes[download_id]


class HttpServerTestContext(object):
    def __init__(self):
        self._server = _TestServer()

    def __exit__(self, type, value, traceback):
        self._server.unlisten()

    def __enter__(self):
        self._server.start()
        return self._server



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
from __future__ import absolute_import

import os

import pytest

from anaconda_project.internal.test.tmpfile_utils import (with_directory_contents, tmp_script_commandline,
                                                          with_directory_contents_completing_project_file)
from anaconda_project.local_state_file import LocalStateFile, DEFAULT_LOCAL_STATE_FILENAME
from anaconda_project.requirements_registry.provider import (Provider, ProvideContext, EnvVarProvider, ProvideResult,
                                                             shutdown_service_run_state)
from anaconda_project.requirements_registry.registry import RequirementsRegistry
from anaconda_project.requirements_registry.requirement import EnvVarRequirement, UserConfigOverrides
from anaconda_project.project import Project
from anaconda_project.provide import PROVIDE_MODE_DEVELOPMENT
from anaconda_project.project_file import DEFAULT_PROJECT_FILENAME
from anaconda_project.prepare import (prepare_without_interaction, unprepare)
from anaconda_project.frontend import NullFrontend
from anaconda_project.test.project_utils import project_no_dedicated_env
from anaconda_project.test.environ_utils import minimal_environ
from anaconda_project.internal import (keyring, conda_api)

conda_env_var = conda_api.conda_prefix_variable()


def test_find_provider_by_class_name():
    registry = RequirementsRegistry()
    found = registry.find_provider_by_class_name(class_name="CondaEnvProvider")
    assert found is not None
    assert found.__class__.__name__ == "CondaEnvProvider"


def test_find_provider_by_class_name_not_found():
    registry = RequirementsRegistry()
    with pytest.raises(ValueError):
        registry.find_provider_by_class_name(class_name="NotAThing")


def test_provider_default_method_implementations():
    class UselessProvider(Provider):
        @property
        def title(self):
            return ""

        def read_config(self, requirement, environ, local_state_file, default_env_spec_name, overrides):
            return dict()

        def provide(self, requirement, context):
            pass

        def unprovide(self, requirement, local_state_file, requirement_status=None):
            pass

        def missing_env_vars_to_configure(self, requirement, environ, local_state_file):
            return ()

        def missing_env_vars_to_provide(self, requirement, environ, local_state_file):
            return ()

    provider = UselessProvider()
    # this method is supposed to do nothing by default (ignore
    # unknown names, in particular)
    provider.set_config_values_as_strings(
        requirement=None,
        environ=None,
        local_state_file=None,
        default_env_spec_name='default',
        overrides=None,
        values=dict())


def _load_env_var_requirement(dirname, env_var):
    project = Project(dirname)

    for requirement in project.requirements(project.default_env_spec_name):
        if isinstance(requirement, EnvVarRequirement) and requirement.env_var == env_var:
            return requirement
    assert [] == project.problems
    raise RuntimeError("No requirement for %s was in the project file, only %r" % (env_var, project.requirements))


def test_env_var_provider_with_no_value():
    def check_env_var_provider(dirname):
        provider = EnvVarProvider()
        requirement = _load_env_var_requirement(dirname, "FOO")
        local_state_file = LocalStateFile.load_for_directory(dirname)
        status = requirement.check_status(dict(), local_state_file, 'default', UserConfigOverrides())
        context = ProvideContext(
            environ=dict(),
            local_state_file=local_state_file,
            default_env_spec_name='default',
            status=status,
            mode=PROVIDE_MODE_DEVELOPMENT,
            frontend=NullFrontend())

        provider.provide(requirement, context=context)
        assert 'FOO' not in context.environ

    with_directory_contents_completing_project_file({
        DEFAULT_PROJECT_FILENAME: """
variables:
  - FOO
"""
    }, check_env_var_provider)


def test_env_var_provider_with_default_value_in_project_file():
    def check_env_var_provider(dirname):
        provider = EnvVarProvider()
        requirement = _load_env_var_requirement(dirname, "FOO")
        assert dict(default='from_default') == requirement.options
        local_state_file = LocalStateFile.load_for_directory(dirname)
        status = requirement.check_status(dict(), local_state_file, 'default', UserConfigOverrides())
        context = ProvideContext(
            environ=dict(),
            local_state_file=local_state_file,
            default_env_spec_name='default',
            status=status,
            mode=PROVIDE_MODE_DEVELOPMENT,
            frontend=NullFrontend())
        result = provider.provide(requirement, context=context)
        assert [] == result.errors
        assert 'FOO' in context.environ
        assert 'from_default' == context.environ['FOO']

    with_directory_contents_completing_project_file(
        {
            DEFAULT_PROJECT_FILENAME: """
variables:
  FOO:
    default: from_default
"""
        }, check_env_var_provider)


def test_env_var_provider_with_unencrypted_default_value_in_project_file_for_encrypted_requirement():
    # the idea here is that if you want to put an unencrypted
    # password in the file, we aren't going to be annoying and
    # stop you.
    def check_env_var_provider(dirname):
        provider = EnvVarProvider()
        requirement = _load_env_var_requirement(dirname, "FOO_SECRET")
        assert requirement.encrypted
        assert dict(default='from_default') == requirement.options
        local_state_file = LocalStateFile.load_for_directory(dirname)
        environ = dict()
        status = requirement.check_status(environ, local_state_file, 'default', UserConfigOverrides())
        context = ProvideContext(
            environ=environ,
            local_state_file=local_state_file,
            default_env_spec_name='default',
            status=status,
            mode=PROVIDE_MODE_DEVELOPMENT,
            frontend=NullFrontend())
        result = provider.provide(requirement, context=context)
        assert [] == result.errors
        assert 'FOO_SECRET' in context.environ
        assert 'from_default' == context.environ['FOO_SECRET']

    with_directory_contents_completing_project_file(
        {
            DEFAULT_PROJECT_FILENAME: """
variables:
  FOO_SECRET:
    default: from_default
"""
        }, check_env_var_provider)


def test_env_var_provider_with_value_set_in_environment():
    def check_env_var_provider(dirname):
        provider = EnvVarProvider()
        requirement = _load_env_var_requirement(dirname, "FOO")
        local_state_file = LocalStateFile.load_for_directory(dirname)
        environ = dict(FOO='from_environ')
        status = requirement.check_status(environ, local_state_file, 'default', UserConfigOverrides())
        assert dict(source='environ', value='from_environ') == status.analysis.config
        context = ProvideContext(
            environ=environ,
            local_state_file=local_state_file,
            default_env_spec_name='default',
            status=status,
            mode=PROVIDE_MODE_DEVELOPMENT,
            frontend=NullFrontend())
        result = provider.provide(requirement, context=context)
        assert [] == result.errors
        assert 'FOO' in context.environ
        assert 'from_environ' == context.environ['FOO']

    # set a default to be sure we prefer 'environ' instead
    with_directory_contents_completing_project_file(
        {
            DEFAULT_PROJECT_FILENAME: """
variables:
  FOO:
    default: from_default
"""
        }, check_env_var_provider)


def test_env_var_provider_with_value_set_in_local_state():
    def check_env_var_provider(dirname):
        provider = EnvVarProvider()
        requirement = _load_env_var_requirement(dirname, "FOO")
        local_state_file = LocalStateFile.load_for_directory(dirname)
        # set in environ to be sure we override it with local state
        environ = dict(FOO='from_environ')
        status = requirement.check_status(environ, local_state_file, 'default', UserConfigOverrides())
        assert dict(value="from_local_state", source="variables") == status.analysis.config
        context = ProvideContext(
            environ=environ,
            local_state_file=local_state_file,
            default_env_spec_name='default',
            status=status,
            mode=PROVIDE_MODE_DEVELOPMENT,
            frontend=NullFrontend())
        result = provider.provide(requirement, context=context)
        assert [] == result.errors
        assert 'FOO' in context.environ
        assert 'from_local_state' == context.environ['FOO']

    with_directory_contents_completing_project_file(
        {
            DEFAULT_PROJECT_FILENAME: """
variables:
  FOO:
    default: from_default
    """,
            DEFAULT_LOCAL_STATE_FILENAME: """
variables:
  FOO: from_local_state
"""
        }, check_env_var_provider)


def test_env_var_provider_with_encrypted_value_set_in_local_state():
    def check_env_var_provider(dirname):
        provider = EnvVarProvider()
        requirement = _load_env_var_requirement(dirname, "FOO_PASSWORD")
        assert requirement.encrypted
        local_state_file = LocalStateFile.load_for_directory(dirname)
        # set in environ to be sure we override it with local state
        environ = dict(FOO_PASSWORD='from_environ')
        status = requirement.check_status(environ, local_state_file, 'default', UserConfigOverrides())
        assert dict(value="from_local_state", source="variables") == status.analysis.config
        context = ProvideContext(
            environ=environ,
            local_state_file=local_state_file,
            default_env_spec_name='default',
            status=status,
            mode=PROVIDE_MODE_DEVELOPMENT,
            frontend=NullFrontend())
        result = provider.provide(requirement, context=context)
        assert [] == result.errors
        assert 'FOO_PASSWORD' in context.environ
        assert 'from_local_state' == context.environ['FOO_PASSWORD']

    with_directory_contents_completing_project_file(
        {
            DEFAULT_PROJECT_FILENAME: """
variables:
  FOO_PASSWORD:
    default: from_default
    """,
            DEFAULT_LOCAL_STATE_FILENAME: """
variables:
  FOO_PASSWORD: from_local_state
"""
        }, check_env_var_provider)


def test_env_var_provider_configure_local_state_value():
    def check_env_var_provider_config_local_state(dirname):
        provider = EnvVarProvider()
        requirement = _load_env_var_requirement(dirname, "FOO")
        local_state_file = LocalStateFile.load_for_directory(dirname)
        status = requirement.check_status(dict(), local_state_file, 'default', UserConfigOverrides())
        assert dict(source='unset') == status.analysis.config

        assert local_state_file.get_value(['variables', 'FOO']) is None

        environ = dict()

        provider.set_config_values_as_strings(requirement, environ, local_state_file, 'default', UserConfigOverrides(),
                                              dict(value="bar"))

        assert local_state_file.get_value(['variables', 'FOO']) == "bar"
        local_state_file.save()

        local_state_file_2 = LocalStateFile.load_for_directory(dirname)
        assert local_state_file_2.get_value(['variables', 'FOO']) == "bar"

        # setting empty string = unset
        provider.set_config_values_as_strings(requirement, environ, local_state_file, 'default', UserConfigOverrides(),
                                              dict(value=""))
        assert local_state_file.get_value(['variables', 'FOO']) is None

        local_state_file.save()

        local_state_file_3 = LocalStateFile.load_for_directory(dirname)
        assert local_state_file_3.get_value(['variables', 'FOO']) is None

    with_directory_contents_completing_project_file({
        DEFAULT_PROJECT_FILENAME: """
variables:
  - FOO
"""
    }, check_env_var_provider_config_local_state)


def test_env_var_provider_configure_encrypted_value():
    def check_env_var_provider_config_local_state(dirname):
        provider = EnvVarProvider()
        requirement = _load_env_var_requirement(dirname, "FOO_PASSWORD")
        assert requirement.encrypted
        local_state_file = LocalStateFile.load_for_directory(dirname)
        status = requirement.check_status(dict(), local_state_file, 'default', UserConfigOverrides())
        assert dict(source='unset') == status.analysis.config

        assert local_state_file.get_value(['variables', 'FOO_PASSWORD']) is None
        assert set(keyring.fallback_data().values()) == set()

        environ = dict(CONDA_DEFAULT_ENV='/pretend/env', CONDA_ENV_PATH='/pretend/env', CONDA_PREFIX='/pretend/env')

        provider.set_config_values_as_strings(requirement, environ, local_state_file, 'default', UserConfigOverrides(),
                                              dict(value="bar"))

        # should not have affected local state, should use keyring
        assert local_state_file.get_value(['variables', 'FOO_PASSWORD']) is None
        assert set(keyring.fallback_data().values()) == set(["bar"])

        # setting empty string = unset
        provider.set_config_values_as_strings(requirement, environ, local_state_file, 'default', UserConfigOverrides(),
                                              dict(value=""))
        assert local_state_file.get_value(['variables', 'FOO_PASSWORD']) is None
        assert set(keyring.fallback_data().values()) == set()

    keyring.enable_fallback_keyring()
    try:
        with_directory_contents_completing_project_file({
            DEFAULT_PROJECT_FILENAME: """
variables:
  - FOO_PASSWORD
"""
        }, check_env_var_provider_config_local_state)
    finally:
        keyring.disable_fallback_keyring()


def test_env_var_provider_configure_disabled_local_state_value():
    def check_env_var_provider_config_disabled_local_state(dirname):
        provider = EnvVarProvider()
        requirement = _load_env_var_requirement(dirname, "FOO")
        local_state_file = LocalStateFile.load_for_directory(dirname)
        status = requirement.check_status(dict(), local_state_file, 'default', UserConfigOverrides())
        assert dict(source='unset') == status.analysis.config

        assert local_state_file.get_value(['variables', 'FOO']) is None
        assert local_state_file.get_value(['disabled_variables', 'FOO']) is None

        environ = dict()

        # source=environ should mean we set disabled_variables instead of variables
        provider.set_config_values_as_strings(requirement, environ, local_state_file, 'default', UserConfigOverrides(),
                                              dict(source='environ', value="bar"))

        assert local_state_file.get_value(['variables', 'FOO']) is None
        assert local_state_file.get_value(['disabled_variables', 'FOO']) == "bar"

        config = provider.read_config(requirement, environ, local_state_file, 'default', UserConfigOverrides())
        assert config == dict(source='unset', value='bar')

    with_directory_contents_completing_project_file({
        DEFAULT_PROJECT_FILENAME: """
variables:
  - FOO
"""
    }, check_env_var_provider_config_disabled_local_state)


def test_env_var_provider_prepare_unprepare():
    def check_env_var_provider_prepare(dirname):
        project = project_no_dedicated_env(dirname)
        result = prepare_without_interaction(project, environ=minimal_environ(FOO='bar'))
        assert result
        status = unprepare(project, result)
        assert status
        assert status.status_description == 'Success.'
        assert project.frontend.logs == [
            "Nothing to clean up for FOO.", ("Current environment is not in %s, no need to delete it." % dirname)
        ]

    with_directory_contents_completing_project_file({
        DEFAULT_PROJECT_FILENAME: """
variables:
  FOO: null
"""
    }, check_env_var_provider_prepare)


def test_provide_context_properties():
    def check_provide_contents(dirname):
        environ = dict(foo='bar')
        local_state_file = LocalStateFile.load_for_directory(dirname)
        requirement = EnvVarRequirement(RequirementsRegistry(), env_var="FOO")
        status = requirement.check_status(environ, local_state_file, 'default', UserConfigOverrides())
        context = ProvideContext(
            environ=environ,
            local_state_file=local_state_file,
            default_env_spec_name='default',
            status=status,
            mode=PROVIDE_MODE_DEVELOPMENT,
            frontend=NullFrontend())
        assert dict(foo='bar') == context.environ
        assert context.status is status

    with_directory_contents(dict(), check_provide_contents)


def test_provide_result_properties():
    empty = ProvideResult.empty()
    assert [] == empty.errors

    full = ProvideResult(['c', 'd'])
    assert ['c', 'd'] == full.errors

    unchanged = full.copy_with_additions()
    assert ['c', 'd'] == unchanged.errors

    unchanged2 = full.copy_with_additions([])
    assert ['c', 'd'] == unchanged2.errors

    extended = full.copy_with_additions(['z'])
    assert ['c', 'd', 'z'] == extended.errors


def test_provide_context_ensure_service_directory():
    def check_provide_contents(dirname):
        environ = dict()
        local_state_file = LocalStateFile.load_for_directory(dirname)
        requirement = EnvVarRequirement(RequirementsRegistry(), env_var="FOO")
        status = requirement.check_status(environ, local_state_file, 'default', UserConfigOverrides())
        context = ProvideContext(
            environ=environ,
            local_state_file=local_state_file,
            default_env_spec_name='default',
            status=status,
            mode=PROVIDE_MODE_DEVELOPMENT,
            frontend=NullFrontend())
        workpath = context.ensure_service_directory("foo")
        assert os.path.isdir(workpath)
        assert workpath.endswith("foo")
        parent = os.path.dirname(workpath)
        assert parent.endswith("services")
        parent = os.path.dirname(parent)
        assert parent == dirname

        # be sure we can create if it already exists
        workpath2 = context.ensure_service_directory("foo")
        assert os.path.isdir(workpath2)
        assert workpath == workpath2

    with_directory_contents(dict(), check_provide_contents)


def test_provide_context_ensure_service_directory_cannot_create(monkeypatch):
    def mock_makedirs(path, mode=0):
        raise IOError("this is not EEXIST")

    monkeypatch.setattr("os.makedirs", mock_makedirs)

    def check_provide_contents(dirname):
        environ = dict()
        local_state_file = LocalStateFile.load_for_directory(dirname)
        requirement = EnvVarRequirement(RequirementsRegistry(), env_var="FOO")
        status = requirement.check_status(environ, local_state_file, 'default', UserConfigOverrides())
        context = ProvideContext(
            environ=environ,
            local_state_file=local_state_file,
            default_env_spec_name='default',
            status=status,
            mode=PROVIDE_MODE_DEVELOPMENT,
            frontend=NullFrontend())
        with pytest.raises(IOError) as excinfo:
            context.ensure_service_directory("foo")
        assert "this is not EEXIST" in repr(excinfo.value)

    with_directory_contents(dict(), check_provide_contents)


def test_provide_context_transform_service_run_state():
    def check_provide_contents(dirname):
        environ = dict()
        local_state_file = LocalStateFile.load_for_directory(dirname)
        local_state_file.set_service_run_state("myservice", dict(port=42))
        requirement = EnvVarRequirement(RequirementsRegistry(), env_var="FOO")
        status = requirement.check_status(environ, local_state_file, 'default', UserConfigOverrides())
        context = ProvideContext(
            environ=environ,
            local_state_file=local_state_file,
            default_env_spec_name='default',
            status=status,
            mode=PROVIDE_MODE_DEVELOPMENT,
            frontend=NullFrontend())

        def transform_it(state):
            assert 42 == state['port']
            state['port'] = 43
            state['foo'] = 'bar'
            return 1234

        result = context.transform_service_run_state("myservice", transform_it)
        assert 1234 == result
        assert dict(port=43, foo='bar') == local_state_file.get_service_run_state("myservice")

    with_directory_contents(dict(), check_provide_contents)


def test_shutdown_service_run_state_nothing_to_do():
    def check(dirname):
        local_state_file = LocalStateFile.load_for_directory(dirname)
        status = shutdown_service_run_state(local_state_file, 'foo')
        assert status
        assert status.status_description == 'Nothing to do to shut down foo.'

    with_directory_contents(dict(), check)


def test_shutdown_service_run_state_command_success():
    def check(dirname):
        local_state_file = LocalStateFile.load_for_directory(dirname)
        true_commandline = tmp_script_commandline("""import sys
sys.exit(0)
""")
        local_state_file.set_service_run_state('FOO', {'shutdown_commands': [true_commandline]})
        status = shutdown_service_run_state(local_state_file, 'FOO')
        assert status
        assert status.status_description == "Successfully shut down FOO."

    with_directory_contents(dict(), check)


def test_shutdown_service_run_state_command_failure():
    def check(dirname):
        local_state_file = LocalStateFile.load_for_directory(dirname)
        false_commandline = tmp_script_commandline("""import sys
sys.exit(1)
""")
        local_state_file.set_service_run_state('FOO', {'shutdown_commands': [false_commandline]})
        status = shutdown_service_run_state(local_state_file, 'FOO')
        assert not status
        assert status.status_description == "Shutdown commands failed for FOO."
        assert status.errors == ["Shutting down FOO, command %r failed with code 1." % false_commandline]

    with_directory_contents(dict(), check)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
"""Windows command line manipulation."""
from __future__ import absolute_import

import subprocess


class WindowsCommandLineException(Exception):
    pass


def windows_split_command_line(command):
    from ctypes import windll, c_int, POINTER, byref
    from ctypes.wintypes import LPCWSTR, LPWSTR, HLOCAL

    CommandLineToArgvW = windll.shell32.CommandLineToArgvW
    CommandLineToArgvW.argtypes = [LPCWSTR, POINTER(c_int)]
    CommandLineToArgvW.restype = POINTER(LPWSTR)

    LocalFree = windll.kernel32.LocalFree
    LocalFree.argtypes = [HLOCAL]
    LocalFree.restype = HLOCAL

    argc = c_int(0)
    argv = CommandLineToArgvW(command, byref(argc))
    if not argv:
        # The docs say CommandLineToArgvW returns NULL on error, but they
        # don't describe any possible errors, so who knows when/if this happens.
        # Maybe only on low memory or something?
        raise WindowsCommandLineException("Windows could not parse command line: " + str(command))
    try:
        result = []
        i = 0
        while i < argc.value:
            try:
                result.append(str(argv[i]))
            except UnicodeEncodeError as e:
                message = (
                    "Windows cannot represent this command line in its character encoding: " + command + ": " + str(e))
                raise WindowsCommandLineException(message)
            i += 1
    finally:
        LocalFree(argv)
    return result


def windows_join_command_line(args):
    """Combine an argv into a Windows command line.

    Note: this can throw on some args you might pass in, but it should not throw
    on args you got from windows_split_command_line, because those are args
    that are representable as a command line. In other words, we throw on args
    that windows_split_command_line won't be able to reproduce.
    """
    if len(args) == 0:
        raise WindowsCommandLineException("Windows has no way to encode an empty arg list as a command line")

    # CommandLineToArgvW treats the first arg specially, which list2cmdline doesn't handle.
    # If the first arg starts with a quote, CommandLineToArgvW returns up to the second quote,
    # without interpreting any escapes. If the first arg doesn't start with a quote, then
    # no quoting/escaping is processed, the first arg is just left alone including embedded
    # quotes until a space (or control char) is encountered. So we can EITHER quote the whole
    # thing (if it contains no quotes) in order to embed spaces; OR don't quote it to embed
    # quotes but not spaces; but there is no way to have both quotes AND spaces in the first
    # arg, as far as I can tell. Also there's no way the first arg can start with a quote.
    if args[0].startswith('"'):
        # note: windows_split_command_line would never give us this args[0], because it would
        # strip off the leading quote
        raise WindowsCommandLineException("Windows does not allow the first arg to start with a quote: " + args[0])
    if '"' in args[0] and ' ' in args[0]:
        # note: windows_split_command_line would never give us this args[0], because it would
        # split on the space and make this two args
        raise WindowsCommandLineException("Windows does not allow the first arg to have both quotes and spaces: " +
                                          args[0])
    first_arg = args[0]
    if ' ' in first_arg:
        first_arg = '"%s"' % first_arg
    return first_arg + ' ' + subprocess.list2cmdline(args[1:])



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2017, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
from __future__ import absolute_import, print_function

import os
import platform
import pytest

import anaconda_project.internal.streaming_popen as streaming_popen
from anaconda_project.internal.py2_compat import _PY2
from anaconda_project.internal.test.tmpfile_utils import tmp_script_commandline


def detect_linesep(lines):
    # We discovered that for Windows / Py 2.7 / PYTHONUNBUFFERED=1
    # the line separator is in fact "\n" for stdout and "\r\n" for stderr
    # so we needed to be more intelligent about detecting this.
    return '\r\n' if any('\r' in line for line in lines) else '\n'


def add_lineseps(lines, sep=None):
    sep = sep or os.linesep
    return list(map(lambda l: l + sep, lines))


def test_streaming():
    print_stuff = tmp_script_commandline(u"""# -*- coding: utf-8 -*-
from __future__ import print_function
import sys
import time

def flush():
    time.sleep(0.05)
    sys.stdout.flush()
    sys.stderr.flush()

print("a")
flush()
print("x", file=sys.stderr)
flush()
print("b")
flush()
print("y", file=sys.stderr)
flush()
print("c")
flush()
print("z", file=sys.stderr)
flush()
print("d")
flush()
# print partial lines with multiple syscalls
for i in [1,2,3,4,5,6]:
  sys.stdout.write("%d" % i)
sys.stdout.write("\\n")
# print unicode stuff, throws exception on Windows
try:
    print("💯 🌟")
    flush()
except Exception:
    print("Windows")
# print many lines at once, and end on non-newline
sys.stdout.write("1\\n2\\n3\\n4\\n5\\n6")
flush()

sys.exit(2)
""")

    stdout_from_callback = []

    def on_stdout(data):
        stdout_from_callback.append(data)

    stderr_from_callback = []

    def on_stderr(data):
        stderr_from_callback.append(data)

    (p, out_lines, err_lines) = streaming_popen.popen(print_stuff, on_stdout, on_stderr)
    sep_out = detect_linesep(out_lines)
    sep_err = detect_linesep(err_lines)

    expected_out = add_lineseps([u'a', u'b', u'c', u'd', u'123456', u'💯 🌟', u'1', u'2', u'3', u'4', u'5'], sep_out)
    if platform.system() == 'Windows':
        # Windows can't output unicode
        if _PY2:
            expected_out[5] = u'\U0001f4af \U0001f31f' + sep_out
        else:
            expected_out[5] = u"Windows" + sep_out

    expected_out.append(u'6')  # no newline after this one
    expected_err = add_lineseps([u'x', u'y', u'z'], sep_err)

    assert expected_out == out_lines
    assert "".join(expected_out) == "".join(stdout_from_callback)

    assert expected_err == err_lines
    assert "".join(expected_err) == "".join(stderr_from_callback)

    assert p.returncode == 2


def test_bad_utf8():
    print_bad = tmp_script_commandline(u"""# -*- coding: utf-8 -*-
from __future__ import print_function
import os
import sys

print("hello")
sys.stdout.flush()
# write some garbage
os.write(sys.stdout.fileno(), b"\\x42\\xff\\xef\\xaa\\x00\\x01\\xcc")
sys.stdout.flush()
print("goodbye")
sys.stdout.flush()

sys.exit(0)
""")

    stdout_from_callback = []

    def on_stdout(data):
        stdout_from_callback.append(data)

    stderr_from_callback = []

    def on_stderr(data):
        stderr_from_callback.append(data)

    (p, out_lines, err_lines) = streaming_popen.popen(print_bad, on_stdout, on_stderr)
    sep_out = detect_linesep(out_lines)

    expected_out = add_lineseps([u'hello', u'B��\x00\x01�goodbye'], sep_out)
    expected_err = []

    assert expected_out == out_lines
    assert expected_err == err_lines

    assert "".join(expected_out) == "".join(stdout_from_callback)
    assert "".join(expected_err) == "".join(stderr_from_callback)


def test_callbacks_are_none():
    print_stuff = tmp_script_commandline(u"""# -*- coding: utf-8 -*-
from __future__ import print_function
import sys

print("a")
print("b", file=sys.stderr)

sys.exit(0)
""")

    (p, out_lines, err_lines) = streaming_popen.popen(print_stuff, None, None)
    sep_out = detect_linesep(out_lines)
    sep_err = detect_linesep(err_lines)

    expected_out = add_lineseps(['a'], sep_out)
    expected_err = add_lineseps(['b'], sep_err)

    assert expected_out == out_lines
    assert expected_err == err_lines

    assert p.returncode == 0


def test_io_error(monkeypatch):
    print_hello = tmp_script_commandline("""from __future__ import print_function
import os
import sys

print("hello")
sys.stdout.flush()

sys.exit(0)
""")

    stdout_from_callback = []

    def on_stdout(data):
        stdout_from_callback.append(data)

    stderr_from_callback = []

    def on_stderr(data):
        stderr_from_callback.append(data)

    def mock_read(*args, **kwargs):
        raise IOError("Nope")

    monkeypatch.setattr("anaconda_project.internal.streaming_popen._read_from_stream", mock_read)

    with pytest.raises(IOError) as excinfo:
        streaming_popen.popen(print_hello, on_stdout, on_stderr)

    assert "Nope" in str(excinfo.value)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
from __future__ import absolute_import, print_function

import copy

from anaconda_project.internal.test.multipart import MultipartEncoder

from tornado import gen
from tornado.httpclient import AsyncHTTPClient, HTTPRequest


@gen.coroutine
def _http_fetch(request):
    http_client = AsyncHTTPClient()
    response = yield http_client.fetch(request)

    if response.error:
        raise response.error
    else:
        raise gen.Return(response)


def http_get_async(url, host=None):
    headers = dict()
    if host is not None:
        headers['Host'] = host
    return _http_fetch(HTTPRequest(url=url, method='GET', headers=headers))


def http_post_async(url, body=None, host=None, headers=None, form=None):
    if form is not None:
        assert body is None
        assert headers is None
        encoder = MultipartEncoder(form)
        body = encoder.to_string()
        headers = {'Content-Type': encoder.content_type}

    if host is not None:
        headers = copy.copy(headers)
        headers['Host'] = host

    return _http_fetch(HTTPRequest(url=url, method='POST', body=body, headers=headers))


def http_get(io_loop, url, host=None):
    return io_loop.run_sync(lambda: http_get_async(url, host))


def http_post(io_loop, url, body=None, host=None, headers=None):
    return io_loop.run_sync(lambda: http_post_async(url, body, host, headers))



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
"""The ``run`` command executes a project, by default without asking questions (fails on missing config)."""
from __future__ import absolute_import, print_function

import sys

from anaconda_project.internal.cli.prepare_with_mode import prepare_with_ui_mode_printing_errors
from anaconda_project.internal.cli.project_load import load_project
from anaconda_project.project_commands import ProjectCommand
from anaconda_project.internal.cli.environment_commands import (create_bootstrap_env, run_on_bootstrap_env)


def _command_from_name(project, command_name):
    command = project.command_for_name(command_name)
    if command is None and command_name is not None:
        # if the command name isn't a configured command name,
        # interpret the command as a notebook or executable.
        attrs = dict(env_spec=project.default_env_spec_name)
        if command_name.lower().endswith(".ipynb"):
            attrs['notebook'] = command_name
        else:
            attrs['args'] = [command_name]

        command = ProjectCommand(name=command_name, attributes=attrs)

    return command


def run_command(project_dir, ui_mode, conda_environment, command_name, extra_command_args):
    """Run the project.

    Returns:
        Does not return if successful.
    """
    project = load_project(project_dir)

    if project.has_bootstrap_env_spec() and not project.is_running_in_bootstrap_env():
        print("Project should be ran by bootstrap env... fixing.")
        create_bootstrap_env(project)
        run_on_bootstrap_env(project)
    else:
        environ = None
        command = _command_from_name(project, command_name)

        result = prepare_with_ui_mode_printing_errors(
            project,
            ui_mode=ui_mode,
            env_spec_name=conda_environment,
            command=command,
            extra_command_args=extra_command_args,
            environ=environ)

        if result.failed:
            # errors were printed already
            return
        elif result.command_exec_info is None:
            print(
                "No known run command for project %s; try adding a 'commands:' section to anaconda-project.yml" %
                project_dir,
                file=sys.stderr)
        else:
            try:

                result.command_exec_info.execvpe()
            except OSError as e:
                print(
                    "Failed to execute '%s': %s" % (" ".join(result.command_exec_info.args), e.strerror),
                    file=sys.stderr)


def main(args):
    """Start the run command and return exit status code.."""
    run_command(args.directory, args.mode, args.env_spec, args.command, args.extra_args_for_command)
    # if we returned, we failed to run the command and should have printed an error
    return 1



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
"""Conda environment providers."""
from __future__ import absolute_import, print_function

import os
import shutil

from anaconda_project.internal import conda_api
from anaconda_project.internal.simple_status import SimpleStatus
from anaconda_project.conda_manager import new_conda_manager, CondaManagerError
from anaconda_project.requirements_registry.provider import EnvVarProvider
from anaconda_project.provide import PROVIDE_MODE_CHECK


def _remove_env_path(env_path):
    """Also used by project_ops.py to delete environment files."""
    if os.path.exists(env_path):
        try:
            shutil.rmtree(env_path)
            return SimpleStatus(success=True, description=("Deleted environment files in %s." % env_path))
        except Exception as e:
            problem = "Failed to remove environment files in {}: {}.".format(env_path, str(e))
            return SimpleStatus(success=False, description=problem)
    else:
        return SimpleStatus(
            success=True, description=("Nothing to clean up for environment '%s'." % os.path.basename(env_path)))


class CondaEnvProvider(EnvVarProvider):
    """Provides a Conda environment."""

    def __init__(self):
        """Override to create our CondaManager."""
        super(CondaEnvProvider, self).__init__()

    def missing_env_vars_to_configure(self, requirement, environ, local_state_file):
        """Override superclass to not require ourselves."""
        return ()

    def missing_env_vars_to_provide(self, requirement, environ, local_state_file):
        """Override superclass to not require ourselves."""
        return self.missing_env_vars_to_configure(requirement, environ, local_state_file)

    def read_config(self, requirement, environ, local_state_file, default_env_spec_name, overrides):
        """Override superclass to add a choice to create a project-scoped environment."""
        assert 'PROJECT_DIR' in environ
        project_dir = environ['PROJECT_DIR']

        if overrides.env_spec_name is not None:
            # short-circuit this whole party
            env = requirement.env_specs.get(overrides.env_spec_name)
            # future: it should be possible to override the env spec without using the
            # default-created project-scoped env.
            config = dict(source='project', env_name=overrides.env_spec_name, value=env.path(project_dir))
            return config

        config = super(CondaEnvProvider, self).read_config(requirement, environ, local_state_file,
                                                           default_env_spec_name, overrides)

        assert 'source' in config

        # we don't support a default here because it would
        # need a hardcoded path which the anaconda-project.yml author
        # would have no way of providing. Fortunately there's
        # no syntax in anaconda-project.yml that should result in setting
        # a default.
        assert config['source'] != 'default'

        if config['source'] == 'unset':
            # if nothing is selected, default to project mode
            # because we don't have a radio button in the UI for
            # "do nothing" right now.
            config['source'] = 'project'

        # if we're supposed to inherit the environment, we don't want to look at
        # anything else. This should always get rid of 'environ' source.
        if local_state_file.get_value('inherit_environment', default=False) and overrides.inherited_env is not None:
            config['source'] = 'inherited'
            config['value'] = overrides.inherited_env

        # convert 'environ' to 'project' when needed... this would
        # happen if you keep the default 'project' choice, so
        # there's nothing in anaconda-project-local.yml
        if config['source'] == 'environ':
            environ_value = config['value']
            project_dir = environ['PROJECT_DIR']
            environ_value_is_project_specific = False
            for env in requirement.env_specs.values():
                if env.path(project_dir) == environ_value:
                    environ_value_is_project_specific = True
            assert environ_value_is_project_specific
            config['source'] = 'project'

        # we should have changed 'environ' to the specific source; since for conda envs
        # we ignore the initial environ value, we always have to track our value in
        assert config['source'] != 'environ'

        # be sure we don't get confused by alternate ways to spell the path
        if 'value' in config:
            config['value'] = os.path.normpath(config['value'])

        config['env_name'] = default_env_spec_name

        if 'value' in config:
            for env in requirement.env_specs.values():
                if config['value'] == env.path(project_dir):
                    config['env_name'] = env.name
                    if config['source'] == 'variables':
                        config['source'] = 'project'
        elif config['source'] == 'project':
            env = requirement.env_specs.get(config['env_name'])
            config['value'] = env.path(project_dir)

        assert 'env_name' in config

        # print("read_config " + repr(config))

        return config

    def set_config_values_as_strings(self, requirement, environ, local_state_file, default_env_spec_name, overrides,
                                     values):
        """Override superclass to support 'project' source option."""
        super(CondaEnvProvider, self).set_config_values_as_strings(requirement, environ, local_state_file,
                                                                   default_env_spec_name, overrides, values)

        # We have to clear out the user override or it will
        # never stop overriding the user's new choice, if they
        # have changed to another env.
        overrides.env_spec_name = None

        if 'source' in values:
            if values['source'] == 'inherited':
                local_state_file.set_value('inherit_environment', True)
                # the superclass should have unset this so we inherit instead of using it
                assert local_state_file.get_value(['variables', requirement.env_var]) is None
            else:
                # don't write this out if it wasn't in there anyway
                if local_state_file.get_value('inherit_environment') is not None:
                    local_state_file.set_value('inherit_environment', False)

            if values['source'] == 'project':
                project_dir = environ['PROJECT_DIR']
                name = values['env_name']
                for env in requirement.env_specs.values():
                    if env.name == name:
                        prefix = env.path(project_dir)
                        local_state_file.set_value(['variables', requirement.env_var], prefix)

    def provide(self, requirement, context):
        """Override superclass to create or update our environment."""
        assert 'PATH' in context.environ

        conda = new_conda_manager(context.frontend)

        # set from the inherited vale if necessary
        if context.status.analysis.config['source'] == 'inherited':
            context.environ[requirement.env_var] = context.status.analysis.config['value']

        # set the env var (but not PATH, etc. to fully activate, that's done below)
        super_result = super(CondaEnvProvider, self).provide(requirement, context)

        project_dir = context.environ['PROJECT_DIR']

        env_name = context.status.analysis.config.get('env_name', context.default_env_spec_name)
        env_spec = requirement.env_specs.get(env_name)

        if context.status.analysis.config['source'] == 'inherited':
            prefix = context.environ.get(requirement.env_var, None)
            inherited = True
        else:
            prefix = None
            inherited = False

        if prefix is None:
            # use the default environment
            prefix = env_spec.path(project_dir)

        assert prefix is not None

        # if the value has changed, choose the matching env spec
        # (something feels wrong here; should this be in read_config?
        # or not at all?)
        for env in requirement.env_specs.values():
            if env.path(project_dir) == prefix:
                env_spec = env
                break

        if context.mode != PROVIDE_MODE_CHECK:
            # we update the environment in both prod and dev mode

            # TODO if not creating a named env, we could use the
            # shared packages, but for now we leave it alone
            assert env_spec is not None
            try:
                conda.fix_environment_deviations(prefix, env_spec, create=(not inherited))
            except CondaManagerError as e:
                return super_result.copy_with_additions(errors=[str(e)])

        conda_api.environ_set_prefix(context.environ, prefix, varname=requirement.env_var)

        path = context.environ.get("PATH", "")

        context.environ["PATH"] = conda_api.set_conda_env_in_path(path, prefix)
        # Some stuff can only be done when a shell is launched:
        #  - we can't set PS1 because it shouldn't be exported.
        #  - we can't run conda activate scripts because they are sourced.
        # We can do these in the output of our activate command, but not here.

        return super_result

    def unprovide(self, requirement, environ, local_state_file, overrides, requirement_status=None):
        """Override superclass to delete project-scoped envs directory."""
        config = self.read_config(
            requirement,
            environ,
            local_state_file,
            # future: pass in this default_env_spec_name
            default_env_spec_name='default',
            overrides=overrides)

        env_path = config.get('value', None)
        assert env_path is not None
        project_dir = environ['PROJECT_DIR']
        if not env_path.startswith(project_dir):
            return SimpleStatus(
                success=True, description=("Current environment is not in %s, no need to delete it." % project_dir))

        return _remove_env_path(env_path)


class CondaBootstrapEnvProvider(EnvVarProvider):
    """Provides a Conda environment."""

    def __init__(self):
        """Override to create our CondaManager."""
        super(CondaBootstrapEnvProvider, self).__init__()

    def missing_env_vars_to_configure(self, requirement, environ, local_state_file):
        """Override superclass to not require ourselves."""
        return ()

    def missing_env_vars_to_provide(self, requirement, environ, local_state_file):
        """Override superclass to not require ourselves."""
        return self.missing_env_vars_to_configure(requirement, environ, local_state_file)

    def read_config(self, requirement, environ, local_state_file, default_env_spec_name, overrides):
        """Override superclass to add a choice to create a project-scoped environment."""
        assert 'PROJECT_DIR' in environ
        project_dir = environ['PROJECT_DIR']

        if overrides.env_spec_name is not None:
            # short-circuit this whole party
            env = requirement.env_specs.get(overrides.env_spec_name)
            # future: it should be possible to override the env spec without using the
            # default-created project-scoped env.
            config = dict(source='project', env_name=overrides.env_spec_name, value=env.path(project_dir))
            return config

        config = super(CondaBootstrapEnvProvider, self).read_config(requirement, environ, local_state_file,
                                                                    default_env_spec_name, overrides)

        assert 'source' in config

        if config['source'] == 'unset':
            # if nothing is selected, default to project mode
            # because we don't have a radio button in the UI for
            # "do nothing" right now.
            config['source'] = 'project'

        # if we're supposed to inherit the environment, we don't want to look at
        # anything else. This should always get rid of 'environ' source.
        if local_state_file.get_value('inherit_environment', default=False) and overrides.inherited_env is not None:
            config['source'] = 'inherited'
            config['value'] = overrides.inherited_env

        # convert 'environ' to 'project' when needed... this would
        # happen if you keep the default 'project' choice, so
        # there's nothing in anaconda-project-local.yml
        if config['source'] == 'environ':
            environ_value = config['value']
            project_dir = environ['PROJECT_DIR']
            environ_value_is_project_specific = False
            for env in requirement.env_specs.values():
                if env.path(project_dir) == environ_value:
                    environ_value_is_project_specific = True
            assert environ_value_is_project_specific
            config['source'] = 'project'

        # we should have changed 'environ' to the specific source; since for conda envs
        # we ignore the initial environ value, we always have to track our value in
        assert config['source'] != 'environ'

        # be sure we don't get confused by alternate ways to spell the path
        if 'value' in config:
            config['value'] = os.path.normpath(config['value'])

        config['env_name'] = default_env_spec_name

        assert config['env_name'] == 'bootstrap-env'

        if 'value' in config:
            for env in requirement.env_specs.values():
                if config['value'] == env.path(project_dir):
                    config['env_name'] = env.name
                    if config['source'] == 'variables':
                        config['source'] = 'project'
        elif config['source'] == 'project':
            env = requirement.env_specs.get(config['env_name'])
            config['value'] = env.path(project_dir)

        assert 'env_name' in config

        return config

    def set_config_values_as_strings(self, requirement, environ, local_state_file, default_env_spec_name, overrides,
                                     values):
        """Override superclass to support 'project' source option."""
        super(CondaBootstrapEnvProvider, self).set_config_values_as_strings(requirement, environ, local_state_file,
                                                                            default_env_spec_name, overrides, values)

        # We have to clear out the user override or it will
        # never stop overriding the user's new choice, if they
        # have changed to another env.
        overrides.env_spec_name = None

        if 'source' in values:
            if values['source'] == 'inherited':
                local_state_file.set_value('inherit_environment', True)
                # the superclass should have unset this so we inherit instead of using it
                assert local_state_file.get_value(['variables', requirement.env_var]) is None
            else:
                # don't write this out if it wasn't in there anyway
                if local_state_file.get_value('inherit_environment') is not None:
                    local_state_file.set_value('inherit_environment', False)

            if values['source'] == 'project':
                project_dir = environ['PROJECT_DIR']
                name = values['env_name']
                for env in requirement.env_specs.values():
                    if env.name == name:
                        prefix = env.path(project_dir)
                        local_state_file.set_value(['variables', requirement.env_var], prefix)

    def provide(self, requirement, context):
        """Override superclass to create or update our environment."""
        assert 'PATH' in context.environ
        conda = new_conda_manager(context.frontend)

        # set from the inherited vale if necessary
        if context.status.analysis.config['source'] == 'inherited':
            context.environ[requirement.env_var] = context.status.analysis.config['value']

        # set the env var (but not PATH, etc. to fully activate, that's done below)
        super_result = super(CondaBootstrapEnvProvider, self).provide(requirement, context)

        project_dir = context.environ['PROJECT_DIR']

        env_name = context.status.analysis.config.get('env_name', context.default_env_spec_name)
        env_spec = requirement.env_specs.get(env_name)

        prefix = os.path.join(project_dir, 'envs', 'bootstrap-env')

        # if the value has changed, choose the matching env spec
        # (something feels wrong here; should this be in read_config?
        # or not at all?)
        for env in requirement.env_specs.values():
            if env.path(project_dir) == prefix:
                env_spec = env
                break

        if context.mode != PROVIDE_MODE_CHECK:
            # we update the environment in both prod and dev mode

            # TODO if not creating a named env, we could use the
            # shared packages, but for now we leave it alone
            assert env_spec is not None
            try:
                conda.fix_environment_deviations(prefix, env_spec, create=True)
            except CondaManagerError as e:
                return super_result.copy_with_additions(errors=[str(e)])

        conda_api.environ_set_prefix(context.environ, prefix, varname=requirement.env_var)

        path = context.environ.get("PATH", "")

        context.environ["PATH"] = conda_api.set_conda_env_in_path(path, prefix)
        # Some stuff can only be done when a shell is launched:
        #  - we can't set PS1 because it shouldn't be exported.
        #  - we can't run conda activate scripts because they are sourced.
        # We can do these in the output of our activate command, but not here.

        return super_result

    def unprovide(self, requirement, environ, local_state_file, overrides, requirement_status=None):
        """Override superclass to delete project-scoped envs directory."""
        config = self.read_config(
            requirement,
            environ,
            local_state_file,
            # future: pass in this default_env_spec_name
            default_env_spec_name='bootstrap-env',
            overrides=overrides)

        env_path = config.get('value', None)
        assert env_path is not None
        project_dir = environ['PROJECT_DIR']
        if not env_path.startswith(project_dir):
            return SimpleStatus(
                success=True, description=("Current environment is not in %s, no need to delete it." % project_dir))

        return _remove_env_path(env_path)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
"""Redis-related requirements."""

from anaconda_project.requirements_registry.requirements.service import ServiceRequirement
# don't "import from" network_util or we can't monkeypatch it in tests
import anaconda_project.requirements_registry.network_util as network_util


class RedisRequirement(ServiceRequirement):
    """A requirement for REDIS_URL (or another specified env var) to point to a running Redis."""

    @property
    def description(self):
        """Override superclass to supply our description."""
        return self._description("A running Redis server, located by a redis: URL set as %s." % (self.env_var))

    def _why_not_provided(self, environ):
        url = self._get_value_of_env_var(environ)
        if url is None:
            return self._unset_message()
        split = network_util.urlparse.urlsplit(url)
        if split.scheme != 'redis':
            return "{env_var} value '{url}' does not have 'redis:' scheme.".format(env_var=self.env_var, url=url)
        port = 6379
        if split.port is not None:
            port = split.port
        if network_util.can_connect_to_socket(split.hostname, port):
            return None
        else:
            return "Cannot connect to Redis at {url}.".format(url=url, env_var=self.env_var)

    def check_status(self, environ, local_state_file, default_env_spec_name, overrides, latest_provide_result=None):
        """Override superclass to get our status."""
        why_not_provided = self._why_not_provided(environ)

        has_been_provided = why_not_provided is None
        if has_been_provided:
            status_description = ("Using Redis server at %s" % self._get_value_of_env_var(environ))
        else:
            status_description = why_not_provided

        return self._create_status(
            environ,
            local_state_file,
            default_env_spec_name,
            overrides=overrides,
            has_been_provided=has_been_provided,
            status_description=status_description,
            provider_class_name='RedisProvider',
            latest_provide_result=latest_provide_result)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
from __future__ import absolute_import, print_function

try:
    from shlex import quote
except ImportError:
    from pipes import quote

import platform
import re

from anaconda_project.internal.cli.main import _parse_args_and_run_subcommand
from anaconda_project.internal.cli.activate import activate, main
from anaconda_project.internal.cli.prepare_with_mode import UI_MODE_TEXT_ASSUME_YES_DEVELOPMENT
from anaconda_project.internal.test.tmpfile_utils import with_directory_contents_completing_project_file
from anaconda_project.project_file import DEFAULT_PROJECT_FILENAME
from anaconda_project.local_state_file import DEFAULT_LOCAL_STATE_FILENAME
from anaconda_project.test.project_utils import project_dir_disable_dedicated_env


class Args(object):
    def __init__(self, **kwargs):
        self.directory = "."
        self.env_spec = None
        self.mode = UI_MODE_TEXT_ASSUME_YES_DEVELOPMENT
        self.command = None
        for key in kwargs:
            setattr(self, key, kwargs[key])


def _monkeypatch_can_connect_to_socket_to_succeed(monkeypatch):
    can_connect_args = dict()

    def mock_can_connect_to_socket(host, port, timeout_seconds=0.5):
        can_connect_args['host'] = host
        can_connect_args['port'] = port
        can_connect_args['timeout_seconds'] = timeout_seconds
        return True

    monkeypatch.setattr("anaconda_project.requirements_registry.network_util.can_connect_to_socket",
                        mock_can_connect_to_socket)

    return can_connect_args


def _filter_activate_result(result, vars):
    additional_output = False
    outputs = {}
    for line in (result or ()):
        match = re.match(r'^export ([^=]*)=(.*)$', line)
        if match:
            var, value = match.groups()
            if var == 'PATH' and platform.system() == 'Windows':
                print("activate changed PATH on Windows and ideally it would not.")
            outputs[var] = value
        else:
            additional_output = True
    if set(outputs) != set(vars) or additional_output:
        import os
        print("os.environ=" + repr(os.environ))
        print("result=" + repr(result))
    return tuple(outputs.get(v) for v in vars)


def test_activate(monkeypatch):
    can_connect_args = _monkeypatch_can_connect_to_socket_to_succeed(monkeypatch)

    def activate_redis_url(dirname):
        project_dir_disable_dedicated_env(dirname)
        result = activate(dirname, UI_MODE_TEXT_ASSUME_YES_DEVELOPMENT, conda_environment=None, command_name=None)
        assert can_connect_args['port'] == 6379
        result = _filter_activate_result(result, ('PROJECT_DIR', 'REDIS_URL'))
        assert result == (quote(dirname), 'redis://localhost:6379')

    with_directory_contents_completing_project_file({
        DEFAULT_PROJECT_FILENAME: """
services:
  REDIS_URL: redis
    """
    }, activate_redis_url)


def test_activate_quoting(monkeypatch):
    def activate_foo(dirname):
        project_dir_disable_dedicated_env(dirname)
        result = activate(dirname, UI_MODE_TEXT_ASSUME_YES_DEVELOPMENT, conda_environment=None, command_name=None)
        result = _filter_activate_result(result, ('FOO', 'PROJECT_DIR'))
        assert result == ("'$! boo'", quote(dirname))

    with_directory_contents_completing_project_file({
        DEFAULT_PROJECT_FILENAME: """
variables:
  FOO: {}
    """,
        DEFAULT_LOCAL_STATE_FILENAME: """
variables:
  FOO: $! boo
"""
    }, activate_foo)


def test_main(monkeypatch, capsys):
    def mock_conda_create(prefix, pkgs, channels):
        raise RuntimeError("this test should not create an environment in %s with pkgs %r" % (prefix, pkgs))

    monkeypatch.setattr('anaconda_project.internal.conda_api.create', mock_conda_create)

    can_connect_args = _monkeypatch_can_connect_to_socket_to_succeed(monkeypatch)

    def main_redis_url(dirname):
        project_dir_disable_dedicated_env(dirname)
        main(Args(directory=dirname))

    with_directory_contents_completing_project_file({
        DEFAULT_PROJECT_FILENAME: """
services:
  REDIS_URL: redis
"""
    }, main_redis_url)

    assert can_connect_args['port'] == 6379

    out, err = capsys.readouterr()

    assert "export REDIS_URL=redis://localhost:6379\n" in out
    assert "" == err


def test_main_dirname_not_provided_use_pwd(monkeypatch, capsys):
    can_connect_args = _monkeypatch_can_connect_to_socket_to_succeed(monkeypatch)

    def main_redis_url(dirname):
        from os.path import abspath as real_abspath

        def mock_abspath(path):
            if path == ".":
                return dirname
            else:
                return real_abspath(path)

        monkeypatch.setattr('os.path.abspath', mock_abspath)
        project_dir_disable_dedicated_env(dirname)
        code = _parse_args_and_run_subcommand(['anaconda-project', 'activate'])
        assert code == 0

    with_directory_contents_completing_project_file({
        DEFAULT_PROJECT_FILENAME: """
services:
  REDIS_URL: redis
"""
    }, main_redis_url)

    assert can_connect_args['port'] == 6379

    out, err = capsys.readouterr()
    assert "export PROJECT_DIR" in out
    assert "export REDIS_URL=redis://localhost:6379\n" in out
    assert "" == err


def test_main_dirname_provided_use_it(monkeypatch, capsys):
    can_connect_args = _monkeypatch_can_connect_to_socket_to_succeed(monkeypatch)

    def main_redis_url(dirname):
        project_dir_disable_dedicated_env(dirname)
        code = _parse_args_and_run_subcommand(['anaconda-project', 'activate', '--directory', dirname])
        assert code == 0

    with_directory_contents_completing_project_file({
        DEFAULT_PROJECT_FILENAME: """
services:
  REDIS_URL: redis
"""
    }, main_redis_url)

    assert can_connect_args['port'] == 6379

    out, err = capsys.readouterr()
    assert "export PROJECT_DIR" in out
    assert "export REDIS_URL=redis://localhost:6379\n" in out
    assert "" == err


def test_main_bad_command_provided(capsys):
    def check(dirname):
        project_dir_disable_dedicated_env(dirname)
        code = _parse_args_and_run_subcommand(
            ['anaconda-project', 'activate', '--directory', dirname, '--command', 'nope'])
        assert code == 1

    with_directory_contents_completing_project_file({
        DEFAULT_PROJECT_FILENAME: """
services:
  REDIS_URL: redis
"""
    }, check)

    out, err = capsys.readouterr()
    assert err.startswith("Command name 'nope' is not in")


def _monkeypatch_can_connect_to_socket_to_fail_to_find_redis(monkeypatch):
    def mock_can_connect_to_socket(host, port, timeout_seconds=0.5):
        if port == 6379:
            return False  # default Redis not there
        else:
            return True  # can't start a custom Redis here

    monkeypatch.setattr("anaconda_project.requirements_registry.network_util.can_connect_to_socket",
                        mock_can_connect_to_socket)


def test_main_fails_to_redis(monkeypatch, capsys):
    _monkeypatch_can_connect_to_socket_to_fail_to_find_redis(monkeypatch)

    def main_redis_url(dirname):
        project_dir_disable_dedicated_env(dirname)
        code = main(Args(directory=dirname))
        assert 1 == code

    with_directory_contents_completing_project_file({
        DEFAULT_PROJECT_FILENAME: """
services:
  REDIS_URL: redis
"""
    }, main_redis_url)

    out, err = capsys.readouterr()
    assert "missing requirement" in err
    assert "All ports from 6380 to 6449 were in use" in err



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
import codecs
import os

import pytest

from anaconda_project.internal.test.tmpfile_utils import with_directory_contents
from anaconda_project.local_state_file import (LocalStateFile, DEFAULT_LOCAL_STATE_FILENAME, SERVICE_RUN_STATES_SECTION,
                                               possible_local_state_file_names)


def test_create_missing_local_state_file():
    def create_file(dirname):
        filename = os.path.join(dirname, DEFAULT_LOCAL_STATE_FILENAME)
        assert not os.path.exists(filename)
        local_state_file = LocalStateFile.load_for_directory(dirname)
        assert local_state_file is not None
        assert not os.path.exists(filename)
        local_state_file.save()
        assert os.path.exists(filename)
        with codecs.open(filename, 'r', 'utf-8') as file:
            contents = file.read()
            # this is sort of annoying that the default empty file
            # has {} in it, but in our real usage we should only
            # save the file if we set something in it probably.
            assert "# Anaconda local project state (specific to this user/machine)\n{}\n" == contents

    with_directory_contents(dict(), create_file)


def _use_existing_local_state_file(relative_name):
    def check_file(dirname):
        filename = os.path.join(dirname, relative_name)
        assert os.path.exists(filename)
        local_state_file = LocalStateFile.load_for_directory(dirname)
        state = local_state_file.get_service_run_state("foobar")
        assert dict(port=42, shutdown_commands=[["foo"]]) == state

    sample_run_states = SERVICE_RUN_STATES_SECTION + ":\n  foobar: { port: 42, shutdown_commands: [[\"foo\"]] }\n"
    with_directory_contents({relative_name: sample_run_states}, check_file)


def test_use_existing_local_state_file_default_name():
    _use_existing_local_state_file(DEFAULT_LOCAL_STATE_FILENAME)


def test_use_existing_local_state_file_all_names():
    for name in possible_local_state_file_names:
        _use_existing_local_state_file(name)


def test_use_empty_existing_local_state_file():
    def check_file(dirname):
        filename = os.path.join(dirname, DEFAULT_LOCAL_STATE_FILENAME)
        assert os.path.exists(filename)
        local_state_file = LocalStateFile.load_for_directory(dirname)
        state = local_state_file.get_service_run_state("foobar")
        assert dict() == state

    with_directory_contents({DEFAULT_LOCAL_STATE_FILENAME: ""}, check_file)


def test_modify_run_state():
    def check_file(dirname):
        filename = os.path.join(dirname, DEFAULT_LOCAL_STATE_FILENAME)
        assert os.path.exists(filename)
        local_state_file = LocalStateFile.load_for_directory(dirname)
        state = local_state_file.get_service_run_state("foobar")
        assert dict(port=42, shutdown_commands=[["foo"]]) == state
        local_state_file.set_service_run_state("foobar", dict(port=43, shutdown_commands=[]))
        local_state_file.save()
        changed = local_state_file.get_service_run_state("foobar")
        assert dict(port=43, shutdown_commands=[]) == changed

        # and we can reload it from scratch
        local_state_file2 = LocalStateFile.load_for_directory(dirname)
        changed2 = local_state_file2.get_service_run_state("foobar")
        assert dict(port=43, shutdown_commands=[]) == changed2

    sample_run_states = SERVICE_RUN_STATES_SECTION + ":\n  foobar: { port: 42, shutdown_commands: [[\"foo\"]] }\n"
    with_directory_contents({DEFAULT_LOCAL_STATE_FILENAME: sample_run_states}, check_file)


def test_get_all_run_states():
    def check_file(dirname):
        filename = os.path.join(dirname, DEFAULT_LOCAL_STATE_FILENAME)
        assert os.path.exists(filename)
        local_state_file = LocalStateFile.load_for_directory(dirname)
        state = local_state_file.get_service_run_state("foo")
        assert dict(port=42) == state
        state = local_state_file.get_service_run_state("bar")
        assert dict(port=43) == state
        states = local_state_file.get_all_service_run_states()
        assert dict(foo=dict(port=42), bar=dict(port=43)) == states

    sample_run_states = SERVICE_RUN_STATES_SECTION + ":\n  foo: { port: 42 }\n  bar: { port: 43 }\n"
    with_directory_contents({DEFAULT_LOCAL_STATE_FILENAME: sample_run_states}, check_file)


def test_run_state_must_be_dict():
    def check_cannot_use_non_dict(dirname):
        local_state_file = LocalStateFile.load_for_directory(dirname)
        with pytest.raises(ValueError) as excinfo:
            local_state_file.set_service_run_state("foo", 42)
        assert "service state should be a dict" in repr(excinfo.value)

    with_directory_contents(dict(), check_cannot_use_non_dict)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
from __future__ import absolute_import, print_function

import os
import platform

from anaconda_project.test.project_utils import project_no_dedicated_env
from anaconda_project.test.environ_utils import minimal_environ, strip_environ
from anaconda_project.internal.cli.main import _parse_args_and_run_subcommand
from anaconda_project.project_file import DEFAULT_PROJECT_FILENAME
from anaconda_project.requirements_registry.requirements.redis import RedisRequirement
from anaconda_project.requirements_registry.registry import RequirementsRegistry
from anaconda_project.internal.test.tmpfile_utils import (tmp_script_commandline,
                                                          with_directory_contents_completing_project_file)
from anaconda_project.internal.simple_status import SimpleStatus
from anaconda_project.local_state_file import LocalStateFile


def _monkeypatch_pwd(monkeypatch, dirname):
    from os.path import abspath as real_abspath

    def mock_abspath(path):
        if path == ".":
            return dirname
        else:
            return real_abspath(path)

    monkeypatch.setattr('os.path.abspath', mock_abspath)


def _monkeypatch_add_service(monkeypatch, result):
    def mock_add_service(*args, **kwargs):
        return result

    monkeypatch.setattr("anaconda_project.project_ops.add_service", mock_add_service)


_echo_commandline = tmp_script_commandline("""from __future__ import print_function
import sys
print(" ".join(sys.argv))
sys.exit(0)
""")


def test_add_service(capsys, monkeypatch):
    def check(dirname):
        _monkeypatch_pwd(monkeypatch, dirname)

        status = SimpleStatus(success=True, description='Service added.')
        status.requirement = RedisRequirement(RequirementsRegistry(), env_var='REDIS_URL', options=dict(type='redis'))

        _monkeypatch_add_service(monkeypatch, status)

        code = _parse_args_and_run_subcommand(['anaconda-project', 'add-service', 'redis'])
        assert code == 0

        out, err = capsys.readouterr()
        assert (
            'Service added.\n' + 'Added service redis to the project file, its address will be in REDIS_URL.\n') == out
        assert '' == err

    with_directory_contents_completing_project_file(dict(), check)


def test_add_service_fails(capsys, monkeypatch):
    def check(dirname):
        _monkeypatch_pwd(monkeypatch, dirname)
        _monkeypatch_add_service(monkeypatch, SimpleStatus(success=False, description='Service add FAIL.'))

        code = _parse_args_and_run_subcommand(['anaconda-project', 'add-service', 'redis'])
        assert code == 1

        out, err = capsys.readouterr()
        assert '' == out
        assert 'Service add FAIL.\n' == err

    with_directory_contents_completing_project_file(dict(), check)


def test_remove_service(capsys, monkeypatch):
    def check(dirname):
        _monkeypatch_pwd(monkeypatch, dirname)
        local_state = LocalStateFile.load_for_directory(dirname)
        local_state.set_service_run_state('ABC', {'shutdown_commands': [_echo_commandline + ['"shutting down ABC"']]})
        local_state.set_service_run_state('TEST', {'shutdown_commands': [_echo_commandline + ['"shutting down TEST"']]})
        local_state.save()

        code = _parse_args_and_run_subcommand(['anaconda-project', 'remove-service', 'TEST'])
        assert code == 0

        out, err = capsys.readouterr()
        assert '' == err
        expected_out = ("Removed service 'TEST' from the project file.\n")
        assert expected_out == out

    with_directory_contents_completing_project_file({
        DEFAULT_PROJECT_FILENAME: 'services:\n  ABC: redis\n  TEST: redis'
    }, check)


def test_remove_service_shutdown_fails(capsys, monkeypatch):
    def check(dirname):
        _monkeypatch_pwd(monkeypatch, dirname)
        local_state = LocalStateFile.load_for_directory(dirname)
        false_commandline = tmp_script_commandline("""import sys
sys.exit(1)
""")
        local_state.set_service_run_state('TEST', {'shutdown_commands': [false_commandline]})
        local_state.save()

        code = _parse_args_and_run_subcommand(['anaconda-project', 'remove-service', 'TEST'])
        assert code == 1

        out, err = capsys.readouterr()
        expected_err = ("Shutting down TEST, command %r failed with code 1.\n" +
                        "Shutdown commands failed for TEST.\n") % false_commandline
        assert expected_err == err
        assert '' == out

    with_directory_contents_completing_project_file({DEFAULT_PROJECT_FILENAME: 'services:\n  TEST: redis'}, check)


def test_remove_service_by_type(capsys, monkeypatch):
    def check(dirname):
        _monkeypatch_pwd(monkeypatch, dirname)
        local_state = LocalStateFile.load_for_directory(dirname)
        local_state.set_service_run_state('TEST', {'shutdown_commands': [_echo_commandline + ['"shutting down TEST"']]})
        local_state.save()

        code = _parse_args_and_run_subcommand(['anaconda-project', 'remove-service', 'redis'])
        assert code == 0

        out, err = capsys.readouterr()
        assert '' == err
        expected_out = ("Removed service 'redis' from the project file.\n")
        assert expected_out == out

    with_directory_contents_completing_project_file({DEFAULT_PROJECT_FILENAME: 'services:\n  TEST: redis'}, check)


def test_remove_service_duplicate(capsys, monkeypatch):
    def check(dirname):
        _monkeypatch_pwd(monkeypatch, dirname)
        local_state = LocalStateFile.load_for_directory(dirname)
        local_state.set_service_run_state('ABC', {'shutdown_commands': [_echo_commandline + ['"shutting down ABC"']]})
        local_state.set_service_run_state('TEST', {'shutdown_commands': [_echo_commandline + ['"shutting down TEST"']]})
        local_state.save()

        code = _parse_args_and_run_subcommand(['anaconda-project', 'remove-service', 'redis'])
        assert code == 1

        out, err = capsys.readouterr()
        assert '' == out
        expected_err = ("Conflicting results, found 2 matches, use list-services"
                        " to identify which service you want to remove\n")
        assert expected_err == err

    with_directory_contents_completing_project_file({
        DEFAULT_PROJECT_FILENAME: 'services:\n  ABC: redis\n  TEST: redis'
    }, check)


def test_remove_service_running_redis(monkeypatch):
    # this test will fail if you don't have Redis installed, since
    # it actually starts it.
    if platform.system() == 'Windows':
        print("Cannot start redis-server on Windows")
        return

    from anaconda_project.requirements_registry.network_util import can_connect_to_socket as real_can_connect_to_socket
    from anaconda_project.requirements_registry.providers.test import test_redis_provider

    can_connect_args_list = test_redis_provider._monkeypatch_can_connect_to_socket_on_nonstandard_port_only(
        monkeypatch, real_can_connect_to_socket)

    def start_local_redis(dirname):
        project = project_no_dedicated_env(dirname)
        result = test_redis_provider._prepare_printing_errors(project, environ=minimal_environ())
        assert result

        local_state_file = LocalStateFile.load_for_directory(dirname)
        state = local_state_file.get_service_run_state('REDIS_URL')
        assert 'port' in state
        port = state['port']

        assert dict(
            REDIS_URL=("redis://localhost:" + str(port)),
            PROJECT_DIR=project.directory_path) == strip_environ(result.environ)
        assert len(can_connect_args_list) >= 2

        pidfile = os.path.join(dirname, "services/REDIS_URL/redis.pid")
        logfile = os.path.join(dirname, "services/REDIS_URL/redis.log")

        try:
            print("listing of services")
            print(repr(os.listdir(os.path.dirname(os.path.dirname(pidfile)))))
        except Exception as e:
            print("failed to list services: " + str(e))

        try:
            print("listing of REDIS_URL")
            print(repr(os.listdir(os.path.dirname(pidfile))))
        except Exception as e:
            print("failed to list REDIS_URL: " + str(e))

        assert os.path.exists(pidfile)
        assert os.path.exists(logfile)

        assert real_can_connect_to_socket(host='localhost', port=port)

        # now clean it up
        code = _parse_args_and_run_subcommand(
            ['anaconda-project', 'remove-service', 'REDIS_URL', '--directory', dirname])
        assert code == 0

        assert not os.path.exists(pidfile)
        assert not os.path.exists(os.path.join(dirname, "services"))
        assert not real_can_connect_to_socket(host='localhost', port=port)

        local_state_file.load()
        assert dict() == local_state_file.get_service_run_state("REDIS_URL")

    with_directory_contents_completing_project_file({
        DEFAULT_PROJECT_FILENAME: "services:\n  REDIS_URL: redis"
    }, start_local_redis)


def test_remove_service_missing_variable(capsys, monkeypatch):
    def check(dirname):
        _monkeypatch_pwd(monkeypatch, dirname)

        code = _parse_args_and_run_subcommand(['anaconda-project', 'remove-service', 'TEST'])
        assert code == 1

        out, err = capsys.readouterr()

        assert "Service 'TEST' not found in the project file.\n" == err
        assert '' == out

    with_directory_contents_completing_project_file({DEFAULT_PROJECT_FILENAME: ''}, check)


def _test_service_command_with_project_file_problems(capsys, monkeypatch, command):
    def check(dirname):
        _monkeypatch_pwd(monkeypatch, dirname)

        code = _parse_args_and_run_subcommand(command)
        assert code == 1

        out, err = capsys.readouterr()
        assert '' == out
        assert ('variables section contains wrong value type 42,' + ' should be dict or list of requirements\n' +
                'Unable to load the project.\n') in err

    with_directory_contents_completing_project_file({DEFAULT_PROJECT_FILENAME: "variables:\n  42"}, check)


def test_add_service_with_project_file_problems(capsys, monkeypatch):
    _test_service_command_with_project_file_problems(capsys, monkeypatch, ['anaconda-project', 'add-service', 'redis'])


def test_remove_service_with_project_file_problems(capsys, monkeypatch):
    _test_service_command_with_project_file_problems(capsys, monkeypatch,
                                                     ['anaconda-project', 'remove-service', 'TEST'])


def test_list_service_with_project_file_problems(capsys, monkeypatch):
    _test_service_command_with_project_file_problems(capsys, monkeypatch, ['anaconda-project', 'list-services'])


def test_list_service(capsys, monkeypatch):
    def check_list(dirname):
        _monkeypatch_pwd(monkeypatch, dirname)
        code = _parse_args_and_run_subcommand(['anaconda-project', 'list-services'])
        assert code == 0

        out, err = capsys.readouterr()
        assert err == ''
        expected_out = """
Services for project: {dirname}

Name       Description
====       ===========
REDIS_URL  A running Redis server, located by a redis: URL set as REDIS_URL.
""".format(dirname=dirname).strip() + "\n"
        assert out == expected_out

    with_directory_contents_completing_project_file({
        DEFAULT_PROJECT_FILENAME: "services:\n  REDIS_URL: redis\n"
    }, check_list)


def test_list_service_with_empty_project(capsys, monkeypatch):
    def check_empty(dirname):
        _monkeypatch_pwd(monkeypatch, dirname)
        code = _parse_args_and_run_subcommand(['anaconda-project', 'list-services'])
        assert code == 0

        out, err = capsys.readouterr()
        assert err == ''
        assert out == "No services found for project: {}\n".format(dirname)

    with_directory_contents_completing_project_file({DEFAULT_PROJECT_FILENAME: ""}, check_empty)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2017, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
from __future__ import absolute_import, print_function

import codecs
import subprocess
from threading import Thread

try:
    from queue import Queue
except ImportError:  # pragma: no cover (py2 only)
    from Queue import Queue  # pragma: no cover (py2 only)

from anaconda_project.internal import logged_subprocess


# this function exists to be mocked in tests
def _read_from_stream(stream, count):
    return stream.read(count)


def _read_and_queue_data(pipe, queue):
    try:
        while True:
            # Reading one character at a time is ridiculous, but
            # the problem is we want to immediately display each
            # "." in conda's "Doing stuff....." sort of output. If
            # we read >1 char, Python apparently will call the
            # read() system call more than once trying to fill up
            # our requested buffer, which is undesirable because
            # it prevents showing streaming progress to the user.
            data = _read_from_stream(pipe, 1)
            if len(data) == 0:
                break
            remaining = data
            while len(remaining) > 0:
                (start, sep, end) = remaining.partition('\n')
                if sep == '':
                    # no newline found, send chunk along
                    queue.put((pipe, remaining, None))
                    remaining = ''
                else:
                    # newline found, send line and then
                    # look for another line
                    queue.put((pipe, start + sep, None))
                    remaining = end
        queue.put((pipe, None, None))
    except Exception as e:
        queue.put((pipe, None, e))


def _reader_thread(pipe, queue):
    t = Thread(target=_read_and_queue_data, args=(pipe, queue))
    t.daemon = True
    t.start()
    return t


def _combine_lines(datas):
    combined = []
    for data in datas:
        if len(combined) == 0 or combined[-1].endswith("\n"):
            combined.append(data)
        else:
            combined[-1] = combined[-1] + data
    return combined


def popen(args, stdout_callback, stderr_callback, **kwargs):
    def ignore_line(line):
        pass

    if stdout_callback is None:
        stdout_callback = ignore_line
    if stderr_callback is None:
        stderr_callback = ignore_line

    p = logged_subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    queue = Queue()

    # Create/destroy reader outside of the threads, since there
    # have been threading bugs in related destructors such as TextIOWrapper
    # https://bugs.python.org/issue28387
    #
    # In case the reader isn't thread-safe (likely) we only use it
    # from one thread at a time, even though all creation/deletion
    # is in this main thread, reading is always in the child
    # threads.
    #
    # we use errors=replace primarily because with strict
    # errors, TextIOWrapper can raise an exception
    # "prematurely" (before returning all valid bytes).
    # Arguably replace is nicer anyway for our purposes.
    #
    # We don't close these readers, because it seems to result
    # in a double-close on the underlying file.
    stdout_wrapper = codecs.getreader('utf-8')(p.stdout, errors='replace')
    stderr_wrapper = codecs.getreader('utf-8')(p.stderr, errors='replace')

    stdout_thread = _reader_thread(stdout_wrapper, queue)
    stderr_thread = _reader_thread(stderr_wrapper, queue)

    stdout_buffer = []
    stderr_buffer = []

    first_error = None
    stdout_joined = False
    stderr_joined = False
    while not (queue.empty() and (stdout_joined and stderr_joined)):
        (which, data, error) = queue.get()
        if error is not None and first_error is None:
            first_error = error
        if data is None:
            if which is stdout_wrapper:
                stdout_thread.join()
                stdout_joined = True
                assert not stdout_thread.is_alive()
            else:
                assert which is stderr_wrapper
                stderr_thread.join()
                stderr_joined = True
                assert not stderr_thread.is_alive()
        else:
            if which is stdout_wrapper:
                stdout_callback(data)
                stdout_buffer.append(data)
            else:
                assert which is stderr_wrapper
                stderr_callback(data)
                stderr_buffer.append(data)

    assert queue.empty()

    p.stdout.close()
    p.stderr.close()

    p.wait()

    stdout_buffer = _combine_lines(stdout_buffer)
    stderr_buffer = _combine_lines(stderr_buffer)

    if first_error is not None:
        raise first_error

    return (p, stdout_buffer, stderr_buffer)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
"""The ``prepare`` command configures a project to run, asking the user questions if necessary."""
from __future__ import absolute_import, print_function

import anaconda_project.internal.cli.console_utils as console_utils
from anaconda_project.internal.cli.prepare_with_mode import prepare_with_ui_mode_printing_errors
from anaconda_project.internal.cli.project_load import load_project


def prepare_command(project_dir, ui_mode, conda_environment, command_name):
    """Configure the project to run.

    Returns:
        Prepare result (can be treated as True on success).
    """
    project = load_project(project_dir)
    if console_utils.print_project_problems(project):
        return False
    result = prepare_with_ui_mode_printing_errors(
        project, env_spec_name=conda_environment, ui_mode=ui_mode, command_name=command_name)

    return result


def main(args):
    """Start the prepare command and return exit status code."""
    if prepare_command(args.directory, args.mode, args.env_spec, args.command):
        print("The project is ready to run commands.")
        print("Use `anaconda-project list-commands` to see what's available.")
        return 0
    else:
        return 1



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
"""Commands related to the downloads section."""
from __future__ import absolute_import, print_function

from anaconda_project.internal.cli.project_load import load_project
from anaconda_project import project_ops
from anaconda_project.internal.cli import console_utils
from anaconda_project.prepare import prepare_without_interaction
from anaconda_project.provide import PROVIDE_MODE_CHECK


def add_service(project_dir, env_spec_name, service_type, variable_name):
    """Add an item to the services section."""
    project = load_project(project_dir)
    status = project_ops.add_service(
        project, env_spec_name=env_spec_name, service_type=service_type, variable_name=variable_name)
    if status:
        print(status.status_description)
        print("Added service %s to the project file, its address will be in %s." % (status.requirement.service_type,
                                                                                    status.requirement.env_var))
        return 0
    else:
        console_utils.print_status_errors(status)
        return 1


def remove_service(project_dir, env_spec_name, variable_name):
    """Remove an item from the services section."""
    project = load_project(project_dir)
    # we don't want to print errors during this prepare, remove
    # service can proceed even though the prepare fails.
    with project.null_frontend():
        result = prepare_without_interaction(project, env_spec_name=env_spec_name, mode=PROVIDE_MODE_CHECK)
    status = project_ops.remove_service(
        project, env_spec_name=env_spec_name, variable_name=variable_name, prepare_result=result)
    if status:
        print(status.status_description)
        return 0
    else:
        console_utils.print_status_errors(status)
        return 1


def list_services(project_dir, env_spec_name):
    """List the services listed on the project."""
    project = load_project(project_dir)
    if console_utils.print_project_problems(project):
        return 1

    if project.services(env_spec_name):
        print("Services for project: {}\n".format(project_dir))
        console_utils.print_names_and_descriptions(project.service_requirements(env_spec_name), name_attr='title')
    else:
        print("No services found for project: {}".format(project_dir))
    return 0


def main_add(args):
    """Start the add-service command and return exit status code."""
    return add_service(args.directory, args.env_spec, args.service_type, args.variable)


def main_remove(args):
    """Start the remove-service command and return exit status code."""
    return remove_service(args.directory, args.env_spec, args.variable)


def main_list(args):
    """Start the list the services command and return exit status code."""
    return list_services(args.directory, args.env_spec)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2016, Anaconda, Inc. All rights reserved.
#
# Licensed under the terms of the BSD 3-Clause License.
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------
"""Environment class representing a conda environment."""
from __future__ import absolute_import

import codecs
import difflib
import os
import re

import anaconda_project.internal.conda_api as conda_api
import anaconda_project.internal.pip_api as pip_api
from anaconda_project.internal.py2_compat import is_string

from anaconda_project.yaml_file import _load_string, _save_file, _YAMLError, ryaml


def _combine_keeping_last_duplicate(items1, items2, key_func=None):
    def default_key(item):
        return item

    if key_func is None:
        key_func = default_key
    items2_keys = set([key_func(item) for item in items2])
    combined = list([item for item in items1 if key_func(item) not in items2_keys])
    combined = combined + list(items2)
    return tuple(combined)


def _conda_combine_key(spec):
    parsed = conda_api.parse_spec(spec)
    if parsed is None:
        # this is broken but we complain about it in project.py, carry on here
        return spec
    else:
        return parsed.name


def _pip_combine_key(spec):
    parsed = pip_api.parse_spec(spec)
    if parsed is None:
        # this is broken but we complain about it in project.py, carry on here
        return spec
    else:
        return parsed.name


def _combine_conda_package_lists(first, second):
    return _combine_keeping_last_duplicate(first, second, key_func=_conda_combine_key)


class EnvSpec(object):
    """Represents a set of required conda packages we could potentially instantiate as a Conda environment."""

    def __init__(self,
                 name,
                 conda_packages,
                 channels,
                 pip_packages=(),
                 description=None,
                 inherit_from_names=(),
                 inherit_from=(),
                 platforms=(),
                 lock_set=None):
        """Construct a package set with the given name and packages.

        Args:
            name (str): name of the package set
            conda_packages (list): list of package specs to pass to conda install
            channels (list): list of channel names
            pip_packages (list): list of pip package specs to pass to pip
            description (str or None): one-sentence-ish summary of what this env is
            inherit_from_name (str or None): name of what we inherit from
            inherit_from (EnvSpec or None): pull in packages and channels from
            lock_set (CondaLockSet): locked packages or None
        """
        assert inherit_from_names is not None
        assert inherit_from is not None

        self._name = name
        self._conda_packages = tuple(conda_packages)
        self._channels = tuple(channels)
        self._pip_packages = tuple(pip_packages)
        self._description = description
        self._logical_hash = None
        self._locked_hash = None
        self._import_hash = None
        self._inherit_from_names = inherit_from_names
        self._inherit_from = inherit_from
        self._lock_set = lock_set
        self._platforms = tuple(conda_api.sort_platform_list(platforms))

        # inherit_from must be a subset of inherit_from_names
        # except that we can have an anonymous base env spec for
        # the global packages/channels sections; if there was an
        # error that kept us from creating one of the specs we
        # name as a parent, then self._inherit_from would be a
        # subset rather than equal.
        for name in tuple([spec.name for spec in self._inherit_from]):
            assert name is None or name in self._inherit_from_names

        conda_specs_by_name = dict()
        for spec in self.conda_packages_for_create:
            # we quietly skip invalid specs here and let them fail
            # somewhere we can more easily report an error message.
            parsed = conda_api.parse_spec(spec)
            if parsed is not None:
                conda_specs_by_name[parsed.name] = spec
        self._conda_specs_for_create_by_name = conda_specs_by_name

        name_set = set()
        conda_constrained_packages = []
        for spec in self.conda_packages:
            parsed = conda_api.parse_spec(spec)
            if parsed is not None:
                name_set.add(parsed.name)
                if parsed.conda_constraint is not None or parsed.pip_constraint is not None:
                    conda_constrained_packages.append(spec)
        self._conda_logical_specs_name_set = name_set
        self.conda_constrained_packages = sorted(conda_constrained_packages)

        pip_specs_by_name = dict()
        for spec in self.pip_packages:
            # we quietly skip invalid specs here and let them fail
            # somewhere we can more easily report an error message.
            parsed = pip_api.parse_spec(spec)
            if parsed is not None:
                pip_specs_by_name[parsed.name] = spec
        self._pip_specs_by_name = pip_specs_by_name

    @property
    def name(self):
        """Get name of the package set.

        May be None for the anonymous shared base spec
        (toplevel packages, channels sections).
        """
        return self._name

    @property
    def description(self):
        """Get the description of the environment."""
        if self._description is None:
            return self._name
        else:
            return self._description

    def _compute_hash(self, conda_packages, platforms):
        import hashlib
        m = hashlib.sha1()
        for p in conda_packages:
            m.update(p.encode("utf-8"))
        for p in self.pip_packages:
            m.update(p.encode("utf-8"))
        for c in self.channels:
            m.update(c.encode("utf-8"))
        for p in platforms:
            m.update(p.encode("utf-8"))
        result = m.hexdigest()
        return result

    @property
    def logical_hash(self):
        """Get a hash of our "logical" requirements.

        (Changing logical requirements could change the lock set
        if we recreate it.)

        Order matters (change in order will count as a change).

        """
        if self._logical_hash is None:
            self._logical_hash = self._compute_hash(self.conda_packages, self.platforms)
        return self._logical_hash

    @property
    def locked_hash(self):
        """Get a hash of our locked packages (what we'd pass to conda create).

        This is used to see if we need to reprepare
        environments. Order matters (change in order will count as
        a change).
        """
        if self._locked_hash is None:
            self._locked_hash = self._compute_hash(self.conda_packages_for_create, platforms=())
        return self._locked_hash

    @property
    def import_hash(self):
        """Get a hash of parts of the env spec that can appear in environment.yml.

        This is used to see if we need to re-import the environment.yml, requirements.txt
        or whatever. Those files don't have platform information.
        """
        if self._import_hash is None:
            self._import_hash = self._compute_hash(self.conda_packages, platforms=())
        return self._import_hash

    def _get_inherited(self, public_attr, key_func=None):
        private_attr = '_' + public_attr

        def getter(spec):
            return getattr(spec, private_attr)

        return self._get_inherited_with_getter(getter, key_func=key_func)

    def _get_inherited_with_getter(self, getter, key_func=None):
        def _linearized_ancestors(specs, accumulator):
            for spec in specs:
                if spec not in accumulator:
                    _linearized_ancestors(spec._inherit_from, accumulator)
                    accumulator.append(spec)

        ancestors = []
        _linearized_ancestors([self], ancestors)
        assert ancestors[-1] is self

        to_combine = []
        for spec in ancestors:
            to_combine.append(getter(spec))
        combined = []
        for item in to_combine:
            combined = _combine_keeping_last_duplicate(combined, item, key_func=key_func)
        return combined

    @property
    def conda_packages(self):
        """Get the conda packages to install in the environment as an iterable."""
        return self._get_inherited('conda_packages', _conda_combine_key)

    @property
    def channels(self):
        """Get the channels to install conda packages from."""
        return self._get_inherited('channels')

    @property
    def platforms(self):
        """Get the platforms the environment can be on."""
        return self._get_inherited('platforms')

    @property
    def pip_packages(self):
        """Get the pip packages to install in the environment as an iterable."""
        return self._get_inherited('pip_packages', _pip_combine_key)

    @property
    def conda_package_names_set(self):
        """Conda package names that we require, as a Python set."""
        return self._conda_logical_specs_name_set

    @property
    def conda_package_names_for_create_set(self):
        """Conda package names that we require, as a Python set."""
        return set(self._conda_specs_for_create_by_name.keys())

    @property
    def conda_package_names_constrained_set(self):
        """List of conda package names with version constraints."""
        return set(self._conda_names_constrained)

    @property
    def pip_package_names_set(self):
        """Pip package names that we require, as a Python set."""
        return set(self._pip_specs_by_name.keys())

    @property
    def lock_set(self):
        """Get ``CondaLockSet`` for this env spec."""
        return self._lock_set

    @property
    def conda_packages_for_create(self):
        """Get conda packages (preferring the lock set list if present)."""
        if self._lock_set is not None and self._lock_set.enabled and self._lock_set.supports_current_platform:
            return self._lock_set.package_specs_for_current_platform
        else:
            return self.conda_packages

    def _specs_for_package_names(self, names, mapping):
        specs = []
        for name in names:
            spec = mapping.get(name, None)
            if spec is not None:
                specs.append(spec)
        return specs

    def specs_for_conda_package_names(self, names):
        """Get the full install specs given an iterable of package names."""
        return self._specs_for_package_names(names, self._conda_specs_for_create_by_name)

    def specs_for_pip_package_names(self, names):
        """Get the full install specs given an iterable of package names."""
        return self._specs_for_package_names(names, self._pip_specs_by_name)

    @property
    def inherit_from(self):
        """Env spec that we inherit stuff from."""
        return self._inherit_from

    @property
    def inherit_from_names(self):
        """Env spec names that we inherit stuff from."""
        return self._inherit_from_names

    def path(self, project_dir):
        """The filesystem path to the default conda env containing our packages."""
        envs_path = os.environ.get('ANACONDA_PROJECT_ENVS_PATH', os.path.join(project_dir, "envs"))

        return os.path.join(envs_path, self.name)

    def diff_from(self, old):
        """A string showing the comparison between this env spec and another one."""
        channels_diff = list(difflib.ndiff(old.channels, self.channels))
        conda_diff = list(difflib.ndiff(old.conda_packages, self.conda_packages))
        pip_diff = list(difflib.ndiff(old.pip_packages, self.pip_packages))
        if pip_diff:
            pip_diff = ["  pip:"] + list(map(lambda x: "    " + x, pip_diff))
        if channels_diff:
            channels_diff = ["  channels:"] + list(map(lambda x: "    " + x, channels_diff))
        return "\n".join(channels_diff + conda_diff + pip_diff)

    def diff_only_removes_notebook_or_bokeh(self, old):
        """Check whether the diff is exclusively removing 'bokeh' or 'notebook'.

        This is used for a hack, because we can auto-add 'bokeh' or 'notebook'
        packages when we anaconda-project init, and that alone shouldn't result
        in being out of sync with the environment.yml.
        """
        to_remove = [("- " + r) for r in ("bokeh", "notebook")]

        def filter_context(items):
            return list(filter(lambda line: line.startswith("- ") or line.startswith("+ "), items))

        conda_diff = filter_context(difflib.ndiff(old.conda_packages, self.conda_packages))
        for r in to_remove:
            if r in conda_diff:
                conda_diff.remove(r)

        if len(conda_diff) > 0:
            return False

        channels_diff = filter_context(difflib.ndiff(old.channels, self.channels))
        if len(channels_diff) > 0:
            return False

        pip_diff = filter_context(difflib.ndiff(old.pip_packages, self.pip_packages))
        if len(pip_diff) > 0:
            return False

        return True

    def to_json(self):
        """Get JSON for an anaconda-project.yml env spec section."""
        # Note that we use _conda_packages (only the packages we
        # introduce ourselves) rather than conda_packages
        # (includes inherited packages).
        packages = list(self._conda_packages)
        pip_packages = list(self._pip_packages)
        if pip_packages:
            packages.append(dict(pip=pip_packages))
        channels = list(self._channels)
        platforms = list(self._platforms)

        # this is a gross, roundabout hack to get ryaml dicts that
        # have ordering... OrderedDict doesn't work because the
        # yaml saver saves them as some "!!omap" nonsense. Other
        # than ordering, the formatting isn't even preserved here.
        template_json = ryaml.load(
            "something:\n    description: null\n" + "    packages: []\n" + "    channels: []\n",
            Loader=ryaml.RoundTripLoader)

        if self._description is not None:
            template_json['something']['description'] = self._description
        else:
            del template_json['something']['description']
        template_json['something']['packages'] = packages
        template_json['something']['channels'] = channels

        # usually "platforms" will be global so don't clutter
        # every env spec by default
        if len(platforms) > 0:
            template_json['something']['platforms'] = platforms

        if len(self.inherit_from_names) > 0:
            if len(self.inherit_from_names) == 1:
                names = self.inherit_from_names[0]
            else:
                names = list(self.inherit_from_names)
            template_json['something']['inherit_from'] = names

        return template_json['something']

    def apply_pins(self, prefix):
        """Write the augmented pinned file in the environment."""
        conda_meta_path = os.path.join(prefix, 'conda-meta')
        if not os.path.isdir(conda_meta_path):
            raise RuntimeError('Expected path {} to exist'.format(conda_meta_path))
        fname = os.path.join(conda_meta_path, 'pinned')
        if os.path.exists(fname):
            with open(fname, 'r') as fp:
                old_list = fp.read()
        else:
            old_list = ''
        new_list = '\n'.join(self.conda_constrained_packages)
        if old_list != new_list:
            if new_list:
                with open(fname + '.__ap_new', 'w') as fp:
                    fp.write(new_list)
            if os.path.exists(fname):
                os.rename(fname, fname + '.__ap_orig')
            if new_list:
                os.rename(fname + '.__ap_new', fname)

    def save_environment_yml(self, filename):
        """Save as an environment.yml file."""
        # here we want to flatten the env spec to include all inherited stuff
        packages = list(self.conda_packages)
        pip_packages = list(self.pip_packages)
        if pip_packages:
            packages.append(dict(pip=pip_packages))
        channels = list(self.channels)

        yaml = ryaml.load("name: " "\ndependencies: []\nchannels: []\n", Loader=ryaml.RoundTripLoader)

        assert self.name is not None  # the global anonymous spec can't be saved
        yaml['name'] = self.name
        yaml['dependencies'] = packages
        yaml['channels'] = channels

        _save_file(yaml, filename)


def _load_environment_yml(filename):
    """Load an environment.yml as an EnvSpec, or None if not loaded."""
    try:
        with codecs.open(filename, 'r', 'utf-8') as file:
            contents = file.read()
        yaml = _load_string(contents)
    except (IOError, _YAMLError):
        return None

    name = None
    if 'name' in yaml:
        name = yaml['name']
    if not name:
        if 'prefix' in yaml and yaml['prefix']:
            name = os.path.basename(yaml['prefix'])

    if not name:
        name = os.path.basename(filename)

    # We don't do too much validation here because we end up doing it
    # later if we import this into the project, and then load it from
    # the project file. We will do the import such that we don't end up
    # keeping the new project file if it's messed up.
    #
    # However we do try to avoid crashing on None or type errors here.

    raw_dependencies = yaml.get('dependencies', [])
    if not isinstance(raw_dependencies, list) or len(raw_dependencies) == 0:
        raw_dependencies = yaml.get('packages', [])
        if not isinstance(raw_dependencies, list):
            raw_dependencies = []

    raw_channels = yaml.get('channels', [])
    if not isinstance(raw_channels, list):
        raw_channels = []

    conda_packages = []
    pip_packages = []

    for dep in raw_dependencies:
        if is_string(dep):
            conda_packages.append(dep)
        elif isinstance(dep, dict) and 'pip' in dep and isinstance(dep['pip'], list):
            for pip_dep in dep['pip']:
                if is_string(pip_dep):
                    pip_packages.append(pip_dep)

    channels = []
    for channel in raw_channels:
        if is_string(channel):
            channels.append(channel)

    return EnvSpec(name=name, conda_packages=conda_packages, channels=channels, pip_packages=pip_packages, platforms=())


_requirement_option_re = re.compile(r'^-([-a-zA-Z0-9]+)\s(.*)')


def _load_requirements_txt(filename):
    """Load a requirements.txt as an EnvSpec, or None if not loaded."""
    try:
        with codecs.open(filename, 'r', 'utf-8') as file:
            lines = file.readlines()
    except (IOError, _YAMLError):
        return None

    # We don't do too much validation here because we end up doing it
    # later if we import this into the project, and then load it from
    # the project file. We will do the import such that we don't end up
    # keeping the new project file if it's messed up.
    #
    # However we do try to avoid crashing on None or type errors here.

    packages = []
    for line in lines:
        line = line.strip()
        # note: comments MUST be at start of line, because
        # urls can have a hash mark in them
        if line.startswith("#") or line == '':
            continue
        m = _requirement_option_re.search(line)
        if m is not None:
            option = m.group(1)
            package = m.group(2)
            # '-e' means a URL requirement.  '-r' means to
            # recursively include another file. other options are
            # simply ignored right now, which won't really work
            # out well, but.
            if option == 'e':
                packages.append(package)
            elif option == 'r':
                path = os.path.join(os.path.dirname(filename), package)
                child_spec = _load_requirements_txt(path)
                if child_spec is not None:
                    packages.extend(child_spec.pip_packages)
        else:
            packages.append(line)

    return EnvSpec(name='default', conda_packages=(), channels=(), pip_packages=packages)


def _load_importable(filename):
    if filename.endswith(".txt"):
        return _load_requirements_txt(filename)
    else:
        return _load_environment_yml(filename)


def _find_importable_spec(directory_path):
    filenames = ("environment.yml", "environment.yaml", 'requirements.txt')
    for filename in filenames:
        full = os.path.join(directory_path, filename)
        spec = _load_importable(full)
        if spec is not None:
            return (spec, filename)

    return (None, None)


def _find_out_of_sync_importable_spec(project_specs, directory_path):
    (spec, filename) = _find_importable_spec(directory_path)

    if spec is None:
        return (None, None)

    for existing in project_specs:
        if existing.name == spec.name and \
           existing.import_hash == spec.import_hash:
            return (None, None)

    return (spec, filename)


# Added so it can be mocked
def _default_env_spec_packages():
    return ['anaconda']


def _anaconda_default_env_spec(shared_base_spec):
    if shared_base_spec is None:
        inherit_from = ()
    else:
        inherit_from = (shared_base_spec, )
    return EnvSpec(
        name="default",
        conda_packages=_default_env_spec_packages(),
        channels=[],
        platforms=conda_api.default_platforms_with_current(),
        description="Default environment spec for running commands",
        inherit_from_names=(),
        inherit_from=inherit_from)
